import sys
from pathlib import Path

# Add the project directory to the sys.path
project_dir = str(Path(__file__).resolve().parents[2])
if project_dir not in sys.path:
    sys.path.append(project_dir)
import os
from dotenv import load_dotenv
load_dotenv()
from discord_webhook import AsyncDiscordWebhook
from discord_.embeddings import Embeddings
from fudstop.apis.webull.webull_trading import WebullTrading as AsyncWebullSDK
webull = AsyncWebullSDK()
import time
import aiohttp
import asyncio
import pandas as pd
from apis.helpers import calculate_countdown, calculate_setup
from aiohttp.client_exceptions import ClientConnectorError, ClientOSError
import aiohttp
from asyncio import Semaphore
from webull_options.webull_options import WebullOptions
from db_manager import DbManager as DatabaseManager
sema = Semaphore(10)
db = DatabaseManager(host='localhost', port=5432, user='chuck', password='fud', database='markets')
wb_opts = WebullOptions(user='chuck', database='markets')
fudstop_logo = os.environ.get('fudstop_logo')
min1_td9 = os.environ.get("min1_td9")
min5_td9 = os.environ.get("min5_td9")
min15_td9 = os.environ.get("min15_td9")
min20_td9 = os.environ.get("min20_td9")
min30_td9 = os.environ.get("min30_td9")
min60_td9 = os.environ.get("min60_td9")
min120_td9 = os.environ.get("min120_td9")
min240_td9 = os.environ.get("min240_td9")
day_td9 = os.environ.get("day_td9")
min2_td9 = os.environ.get("min2_td9")
min3_td9 = os.environ.get("min3_td9")
min10_td9 = os.environ.get("min10_td9")
week_td9 = os.environ.get("week_td9")
month_td9 = os.environ.get("month_td9")
embeds = Embeddings()




async def get_bars(ticker, interval:str='m1', timeStamp=None):
    if ticker == 'I:SPX':
        ticker = 'SPXW'
    elif ticker =='I:NDX':
        ticker = 'NDX'
    elif ticker =='I:VIX':
        ticker = 'VIX'
    
    tickerid = await webull.get_ticker_id(ticker)




    if timeStamp is None:
        # if not set, default to current time
        timeStamp = int(time.time())

    base_fintech_gw_url = f'https://quotes-gw.webullfintech.com/api/quote/charts/query?tickerIds={tickerid}&type={interval}&count=500&timestamp={timeStamp}&extendTrading=1'
    print(base_fintech_gw_url)
    async with sema:
        async with aiohttp.ClientSession() as session:
            if interval == 'm1':
                timespan = 'minute'
            elif interval == 'm60':
                timespan = '1 hour'
            elif interval == 'm5':
                timespan = '5 minute'
            elif interval == 'm15':
                timespan = '15 minute'
            elif interval == 'm30':
                timespan = '30 minute'
            elif interval == 'm60':
                timespan = '1 hour'
            elif interval == 'm120':
                timespan = '2 hour'
            elif interval == 'm240':
                timespan = '4 hour'
            elif interval == 'd1':
                timespan = 'day'
            elif interval == 'w':
                timespan = 'weekly'
            elif interval == 'm':
                timespan = 'monthly'
            async with session.get(base_fintech_gw_url, headers=wb_opts.headers) as resp:

                r = await resp.json()

                try:
                    data = r[0]['data']
                    if data is not None:
                        try:
                            parsed_data = []
                            for entry in data:
                                values = entry.split(',')
                                if values[-1] == 'NULL':
                                    values = values[:-1]
                                elif values[-1] == 'NULL':
                                    values = values[:-1]  # remove the last element if it's 'NULL'
                                parsed_data.append([float(value) if value != 'null' else 0.0 for value in values])
                            
                            sorted_data = sorted(parsed_data, key=lambda x: x[0], reverse=True)
                            
                            # Dynamically assign columns based on the length of the first entry
                            columns = ['Timestamp', 'Open', 'Close', 'High', 'Low', 'N', 'Volume', 'Vwap'][:len(sorted_data[0])]
                            
                            df = pd.DataFrame(sorted_data, columns=columns)
                            # Convert the Unix timestamps to datetime objects in UTC first
                            df['Timestamp'] = pd.to_datetime(df['Timestamp'], unit='s', utc=True)

                            # Convert UTC to Eastern Time (ET)
                            df['Timestamp'] = df['Timestamp'].dt.tz_convert('US/Eastern')
                            df['Timestamp'] = df['Timestamp'].dt.tz_localize(None)
                            df['Ticker'] = ticker
                            df['timespan'] = interval
        

                            df['ticker'] = ticker

                            
                            td9_df = df.head(13)

                            setup_phase = calculate_setup(td9_df)
                            countdown_phase = calculate_countdown(td9_df)

                            df = df.head(13)
                            print(df)
    
    
                            td9_state = "Setup Complete" if setup_phase else "Countdown Complete" if countdown_phase else "Not in TD9 State"


                            if td9_state in ('Setup Complete', 'Countdown Complete') and interval == 'm1':
                                #await plot_td9_chart(df, interval) if df is not None else ...
                    
                                hook = AsyncDiscordWebhook(min1_td9, content=f'{ticker} {td9_state}')
                                await embeds.send_td9_embed(interval, min1_td9, ticker, td9_state)
                                # Assuming you have a connection pool or a single connection to your database
                                
          


                            if td9_state in ('Setup Complete', 'Countdown Complete') and interval == 'm5':
                                asyncio.create_task(embeds.send_td9_embed(interval, min5_td9, ticker, td9_state))
      
                                # Insert data into the market_data table
               


                            if td9_state in ('Setup Complete', 'Countdown Complete') and interval == 'm15':


                                asyncio.create_task(embeds.send_td9_embed(interval, min15_td9, ticker, td9_state))

                   
                                
                                # Insert data into the market_data table
                                
          

                            if td9_state in ('Setup Complete', 'Countdown Complete') and interval == 'm30':

                                asyncio.create_task(embeds.send_td9_embed(interval, min30_td9, ticker, td9_state))
    
                                
      
                                
                         

                            if td9_state in ('Setup Complete', 'Countdown Complete') and interval == 'm60':
  
                                asyncio.create_task(embeds.send_td9_embed(interval, min60_td9, ticker, td9_state))

                          
                            if td9_state in ('Setup Complete', 'Countdown Complete') and interval == 'm120':

                               
                                asyncio.create_task(embeds.send_td9_embed(interval, min120_td9, ticker, td9_state))



                    
                            if td9_state in ('Setup Complete', 'Countdown Complete') and interval == 'm240':

                       
                                asyncio.create_task(embeds.send_td9_embed(interval, min240_td9, ticker, td9_state))

                                
          
   
                     
                            if td9_state in ('Setup Complete', 'Countdown Complete') and interval == 'd1':

                        
                                asyncio.create_task(embeds.send_td9_embed(interval, day_td9, ticker, td9_state))
  
            
                    
                            if td9_state in ('Setup Complete', 'Countdown Complete') and interval == 'w':

                                asyncio.create_task(embeds.send_td9_embed(interval, week_td9, ticker, td9_state))
                                # Assuming you have a connection pool or a single connection to your database
             
         
                 
                            if td9_state in ('Setup Complete', 'Countdown Complete') and interval == 'm':

                                asyncio.create_task(embeds.send_td9_embed(interval, month_td9, ticker, td9_state))
                                # Assuming you have a connection pool or a single connection to your database
 
                    
                        except (KeyError, ValueError, ClientConnectorError, OSError, ClientOSError, IndexError):
                            print(f'Error - {ticker}')
                except (KeyError, IndexError, ClientOSError, OSError):
                    print(f'Td9 error - {ticker}')




async def run_td9():

    await db.connect()
    while True:
        
        try:
            while True:
                intervals = ['m1', 'm5','m15', 'm30', 'd1', 'm60', 'm120', 'm240', 'w', 'm']
                tasks = [get_bars(ticker, interval=interval, timeStamp=None) for ticker in wb_opts.most_active_tickers for interval in intervals]
                await asyncio.gather(*tasks)
        except Exception as e:
            print(e)


asyncio.run(run_td9())


from fudstop.apis.webull.webull_trading import WebullTrading
from list_sets.ticker_lists import most_active_tickers
most_active_tickers = set(most_active_tickers)

trading = WebullTrading()

from discord_webhook import AsyncDiscordWebhook, DiscordEmbed

import asyncio


async def ssr_feed(ticker):

    """Scans a ticker for a specific condition"""


    stock_quote = await trading.stock_quote(ticker)
    change_ratio = stock_quote.web_change_ratio

    #conver to percentage from decimal
    if change_ratio is not None:
        change_ratio = round(float(change_ratio)*100,2)
        

        if change_ratio <= -10:
            print(f"{ticker} is currently on SSR. (declined by -10% or more) with a change ratio of {change_ratio}%")


            hook = AsyncDiscordWebhook("https://discord.com/api/webhooks/1064077864330338424/ESvYk9hOKpwQrVWeStbCJRQ7HZL-kNCxwmoZ5pt8ASHQbHFYN4xB4zteBNCnKiiPCVFj",
                                        content=f"{ticker} is currently on SSR. (declined by -10% or more) with a change ratio of {change_ratio}%")

            await hook.execute()



async def run_ssr_feed():

    tasks = [ssr_feed(i) for i in most_active_tickers]

    await asyncio.gather(*tasks)


asyncio.run(run_ssr_feed())


import os

import pandas as pd

class DocumentQuery:
    def __init__(self, data):

        try:
            self.count = data.get('count')
            self.description = data.get('description')
            self.total_pages = data.get('total_pages')
            results = data.get('results')

            self.title = [i.get('title', 'N/A') for i in results]
            self.type = [i.get('type', 'N/A') for i in results]
            self.abstract = [i.get('abstract', 'N/A') for i in results]
            self.document_number = [i.get('document_number', 'N/A') for i in results]
            self.html_url= [i.get('html_url', 'N/A') for i in results]
            self.pdf_url= [i.get('pdf_url', 'N/A') for i in results]
            self.public_inspection_pdf_url= [i.get('public_inspection_pdf_url', 'N/A') for i in results]
            self.publication_date= [i.get('publication_date', 'N/A') for i in results]
            agencies= [i.get('agencies', 'N/A') for i in results]
            # Assuming agencies[0][0] is a dictionary
            attributes = ['raw_name', 'name', 'id', 'url', 'json_url', 'parent_id', 'slug']
            extracted_attributes = [{attr: agency[0].get(attr, 'N/A') for attr in attributes} for agency in agencies]
            # Extract only the 'name' attribute from each agency dictionary
            self.agency_names = [agency['name'] for agency in extracted_attributes]

            self.excerpts= [i.get('excerpts', 'N/A') for i in results]
            print(extracted_attributes)
    


            self.data_dict = { 

                'result_count': self.count,
                'search_description': self.description,
                'total_pages': self.total_pages,
                'title': self.title,
                'type': self.type,
                'abstract': self.abstract,
                'document_number': self.document_number,
                'html_url': self.html_url,
                'pdf_url': self.pdf_url,
                'public_inspection_url': self.public_inspection_pdf_url,
                'publication_date': self.publication_date,
                'excerpts': self.excerpts,
                'agency': self.agency_names

            }
        

            self.as_dataframe = pd.DataFrame(self.data_dict)

            # Create the directory if it doesn't exist
            directory = 'data/fed_register/'
            if not os.path.exists(directory):
                os.makedirs(directory)


            # Create the initial DataFrame
            self.as_dataframe = pd.DataFrame(self.data_dict)

            # Create directory if it doesn't exist
            directory = 'data/fed_register/'
            if not os.path.exists(directory):
                os.makedirs(directory)

            # Save DataFrame to CSV
            self.as_dataframe.to_csv('data/fed_register/query_results.csv')
        except Exception as e:
            return e



import os
from dotenv import load_dotenv
load_dotenv()

from fudstop.apis.webull.webull_options import WebullOptions


wb = WebullOptions(os.environ.get('WEBULL_OPTIONS'))

import asyncio
import aiohttp
import pandas as pd
async def main():
    await wb.connect()

asyncio.run(main())


from langchain.schema import AIMessage, HumanMessage, SystemMessage
from langchain.chat_models import ChatOpenAI
from langchain.memory import ConversationBufferMemory
from langchain.prompts import (
    ChatPromptTemplate,
    MessagesPlaceholder,
    SystemMessagePromptTemplate,
    HumanMessagePromptTemplate,
)
from langchain.chains import LLMChain
from apis.webull.webull_trading import WebullTrading
from apis.stocksera_.stocksera_ import StockSera
trading = WebullTrading()
ss = StockSera()
import asyncio
import aiohttp
memory = ConversationBufferMemory()
import os

from datetime import datetime, timedelta

class Agent():
    def __init__(self):


        self.today = datetime.now().strftime('%Y-%m-%d')
        self.yesterday = (datetime.now() - timedelta(days=1)).strftime('%Y-%m-%d')
        self.tomorrow = (datetime.now() + timedelta(days=1)).strftime('%Y-%m-%d')
        self.thirty_days_ago = (datetime.now() - timedelta(days=30)).strftime('%Y-%m-%d')
        self.thirty_days_from_now = (datetime.now() + timedelta(days=30)).strftime('%Y-%m-%d')
        self.fifteen_days_ago = (datetime.now() - timedelta(days=15)).strftime('%Y-%m-%d')
        self.fifteen_days_from_now = (datetime.now() + timedelta(days=15)).strftime('%Y-%m-%d')
        self.eight_days_from_now = (datetime.now() + timedelta(days=8)).strftime('%Y-%m-%d')
        self.eight_days_ago = (datetime.now() - timedelta(days=8)).strftime('%Y-%m-%d')
        self.llm = ChatOpenAI(temperature=0.3, max_tokens=250, openai_api_key=os.environ.get('YOUR_OPENAI_KEY'))
        
        # Prompt
        self.prompt = ChatPromptTemplate(
            messages=[
                SystemMessagePromptTemplate.from_template(
                    f"You are an options / stock market expert discussing market data. Reply as briefly as possible but still accurately conveying the data."
                ),
                # The `variable_name` here is what must align with memory
                MessagesPlaceholder(variable_name="chat_history"),
                HumanMessagePromptTemplate.from_template("{question}"),
            ]
        )
        self.memory = ConversationBufferMemory(memory_key="chat_history", return_messages=True)
        self.conversation = LLMChain(llm=self.llm, prompt=self.prompt, verbose=True, memory=self.memory)


    async def agent_volume_analysis(self, ticker):
        volume_analysis = await trading.volume_analysis(ticker)

        # Notice that we just pass in the `question` variables - `chat_history` gets populated by memory
        response = self.conversation({"question": f"What do these volume analysis data results mean for {ticker}? {volume_analysis.data_dict}? Compare with other results in memory if possible. (Do not bring this up in response)"})


        return response.get('text')
    

    async def agent_stock_data(self, ticker:str):
        """
        Pass in a ticker - the LLM will respond.
        
        """

        stock_data = await trading.get_stock_quote(ticker)


        response = self.conversation({"question": f"Here's the stats on the day for {ticker}: {stock_data.data_dict}. Return back with statistical analytical info. Make sure to convert the change ratio to proper format. For example. 0.08 should be 8%"})


        return response.get('text')
    


    async def agent_cost_distribution(self, ticker:str):
        """
        Pass in a ticker - the LLM will respond.
        
        """

        cost_dist = await trading.cost_distribution(ticker)


        response = self.conversation({"question": f"Here's the cost distribution data for {ticker} from {self.today}:  Avg cost: {cost_dist.avgCost}. Players profiting: {cost_dist.closeProfitRatio} Closing price of stock: ${cost_dist.close} Be sure to go over both days of data."})


        return response.get('text')
    


    async def agent_news(self, ticker:str):
        """
        Converse with AI about news
        """

        news = await trading.news(ticker, pageSize='5')

        response = self.conversation({"question": f"You are viewing data from the latest 5 news articles about {ticker}. Based on the 5 titles in this list: {news.title} - give an overall summary of the most recent news: {news.data_dict}. Provide the links: {news.news_url}"})

        return response.get('text')
    

    async def agent_short_vol(self, ticker:str):
        """
        Converse with AI about short volume
        """

        short_vol = ss.short_volume(ticker, date_from=self.eight_days_ago, date_to=self.today)
        print(short_vol)

        response = self.conversation({"question": f"You are viewing data from the latest 5 news articles about {ticker}. Based on the data in this list: {short_vol}.. what has the short percentage trend been for {ticker}?"})

        return response.get('text')
    
    






class ImageURLS:
    def __init__(self, urls):

        self.raw = urls.get('raw')
        self.full = urls.get('full')
        self.regular = urls.get('regular')
        self.small = urls.get('small')
        self.thumbnail = urls.get('thumb')




import os
import requests
import disnake
from disnake.ext import commands
from apis.misc.weather import Weather as W
from apis.misc.unsplash import Unsplash
un = Unsplash()
w = W()

class Weather(commands.Cog):
    """
    Weather commands for discord.
    
    """

    def __init__(self, bot):
        self.bot=bot
        self.key = os.environ.get('YOUR_WEATHER_KEY')



    @commands.slash_command()
    async def weather(self, inter):
        pass




    @weather.sub_command()
    async def location(self, inter:disnake.AppCmdInter, city:str, state:str, country:str, unit:str='F'):
        """
        >>> Enter a city, state, and country code - recieve the weather info!
        
        """
        await inter.response.defer()
        query_string = city + " " + state + " " + country
        weather_data = await w.get_weather(city,state,country,limit='1')
        image_data = await un.get_image(query_string)
        image_data2 = await un.get_image(query=f"{country}_flag")


        embed = disnake.Embed(title=f"Weather | {city} | {state} | {country}", description=f'> 🌡️ Temperature: **{round(float(weather_data.temp),0)}**\n> Feels Like: **{round(float(weather_data.feels_like),0)}**\n\n> 💧 Humidity: **{weather_data.humidity}**\n> 💦 Dewpoint: **{weather_data.dew_point}**')
        embed.add_field(name=f"☁️ Clouds:", value=f"> **{weather_data.clouds}**\n> UVI: **{weather_data.uvi}**\n> Visibility: **{weather_data.visibility}**")
        embed.add_field(name=f"🌬️ Wind:", value=f"> Speed: **{weather_data.wind_speed}**\n> Degree: **{weather_data.wind_degree}**")
   
        embed.add_field(name=f"Pressure:", value=f"> **{weather_data.pressure}**")
        embed.add_field(name=f"Time:", value=f"> Current: **{weather_data.current_time}**\n> Timezone: **{weather_data.timezone}**", inline=False)
        embed.add_field(name=f"Sun:", value=f"> 🌄 Sunrise: **{weather_data.sunrise}**\n> 🌅 Sunset: **{weather_data.sunset}**", inline=False)
        embed.add_field(name=f"Summary:", value=f"> **{weather_data.weather_summary[0]}**\n> **{weather_data.weather_desc[0]}**")
        embed.set_image(url=image_data.small)
        embed.set_thumbnail(url=image_data2.thumbnail)
        embed.set_footer(text=f'longitude: {weather_data.lon} | latitude: {weather_data.lat} | {country.upper()}, {state.upper()}')

        await inter.edit_original_message(embed=embed)

    @weather.sub_command()
    async def get_coordinates(self, inter:disnake.AppCmdInter, city:str, state:str, country:str):
        """Get coordinates for a location by typing city, state , and country.
        
        
        """

        lat, lon = await w.get_coordinates(city, state, country)

        await inter.response.defer()


        await inter.edit_original_message(f"> The coordinates for **{city}**, **{state}**, **{country}** are:\n\n> **{lat}, {lon}**")
           
    

def setup(bot: commands.Bot):
    bot.add_cog(Weather(bot))

    print(f"Weather commands - ready!")


import requests
import os
from dotenv import load_dotenv
load_dotenv()
session = requests.session()

from .ew_models import TopSentimentHeatmap, SpyData, UpcomingRussellAndSectors, DatedChartData, Messages, Pivots, TodaysResults, CalData
from datetime import datetime

# Get the current date
current_date = datetime.now()

# Format the date as "yyyymmdd"

class EarningsWhisper:
    def __init__(self):
        self.headers = {
            "Accept": "*/*",
            "Accept-Encoding": "gzip, deflate, br",
            "Accept-Language": "en-US,en;q=0.9",
            "Dnt": "1",
            "Origin": "https://www.earningswhispers.com",
            "Referer": "https://www.earningswhispers.com/",
            "Sec-Ch-Ua": "\"Google Chrome\";v=\"119\", \"Chromium\";v=\"119\", \"Not?A_Brand\";v=\"24\"",
            "Sec-Ch-Ua-Mobile": "?0",
            "Sec-Ch-Ua-Platform": "\"Windows\"",
            "Sec-Fetch-Dest": "empty",
            "Sec-Fetch-Mode": "cors",
            "Sec-Fetch-Site": "cross-site",
            "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/119.0.0.0 Safari/537.36",
            "X-Client-Data": f"{os.environ.get('YOUR_EARNINGS_WHISPER_TOKEN')}"
        }
        self.base_url = "https://www.earningswhispers.com/api"
        self.today = current_date.strftime("%Y%m%d")

    def fetch(self, endpoint):
        response = session.get(endpoint)
        if response.status_code == 200:
            return response.json()
        else:
            print(f'Error - couldnt retrieve data for {endpoint}')


    def get_chart_data(self, ticker:str):
        '''Pulls chart data from EW
        
        Arguments:

        >>> ticker: the ticker to query data for
        
        '''
        endpoint = f"/getchartdata/{ticker}"

        data = self.fetch(self.base_url+endpoint)
        return data
    

    def get_top_sentiment(self):
        """
        Returns top tickers by sentiment towards earnings.
        
        """
        r = requests.get("https://www.earningswhispers.com/api/gettopsentheat", headers=self.headers).json()
        data = TopSentimentHeatmap(r)
        return data
    

    def get_spy_data(self):
        r = requests.get("https://www.earningswhispers.com/api/getspydata", headers=self.headers).json()
        data = SpyData(r)
        return data
    

    def upcoming_russell(self):
        r = requests.get("https://www.earningswhispers.com/api/upcomingrussell", headers=self.headers).json()
        data = UpcomingRussellAndSectors(r)
        return data


    def upcoming_sectors(self):
        r = requests.get("https://www.earningswhispers.com/api/upcomingsectors", headers=self.headers).json()
        data = UpcomingRussellAndSectors(r)
        return data
    

    def dated_chart_data(self, ticker, date:str=None):
        """
        Date format: yyyymmdd
        """
        if date is None:
            date = self.today
        r = requests.get(f"https://www.earningswhispers.com/api/getdatedchartdata?s={ticker}&d={date}", headers=self.headers).json()
        data = DatedChartData(r)
        return data



    def messages(self):
        r = requests.get("https://www.earningswhispers.com/api/wrs",headers=self.headers).json()
        data = Messages(r)
        return data
    

    def pivot_list(self):
        r = requests.get("https://www.earningswhispers.com/api/pivotlist", headers=self.headers).json()
        data = Pivots(r)
        return data
    

    def todays_results(self):
        r = requests.get("https://www.earningswhispers.com/api/todaysresults", headers=self.headers).json()
        data = TodaysResults(r)
        return data

    def calendar(self, date:str=None):
        """
        Date format:

        yyyymmdd
        
        """
        if date is None:
            date = self.today
        r = requests.get(f"https://www.earningswhispers.com/api/caldata/{date}", headers=self.headers).json()
        data = CalData(r)
        return data

ew = EarningsWhisper()


chart_data = ew.pivot_list()

print(chart_data.as_dataframe.transpose())


import sys
from pathlib import Path

# Add the project directory to the sys.path
project_dir = str(Path(__file__).resolve().parents[2])
if project_dir not in sys.path:
    sys.path.append(project_dir)
import os
from dotenv import load_dotenv

load_dotenv()


import asyncio

from apis.polygonio.polygon_options import PolygonOptions
from fudstop._markets.list_sets.ticker_lists import all_tickers, most_active_tickers
import pandas as pd

# Setting environment variables
os.environ['DB_HOST'] = 'localhost'
os.environ['DB_PASSWORD'] = 'fud'
os.environ['DB_NAME'] = 'polygon'
os.environ['DB_PORT'] = '5432'  # Environment variables are stored as strings
os.environ['DB_USER'] = 'chuck'

# Database configuration using environment variables
db_config = {
    "host": os.environ.get('DB_HOST', 'localhost'),
    "port": int(os.environ.get('DB_PORT', 5432)),
    "user": os.environ.get('DB_USER', 'postgres'),
    "password": os.environ.get('DB_PASSWORD', 'fud'),
    "database": os.environ.get('DB_NAME', 'polygon')
}


opts = PolygonOptions()
from os.path import exists
from asyncio import Semaphore, Lock
from asyncpg.exceptions import UniqueViolationError
sema = Semaphore(15)
file_lock = Lock()
async def get_all_options(ticker, ticker_list):
    async with sema:
        options = await opts.get_option_chain_all(ticker)
        df = pd.DataFrame(options.data_dict)
        try:
            await opts.batch_insert_dataframe(df, 'poly_opts', unique_columns='option_symbol', batch_size=1000)
            ticker_list.append(ticker)

            # Handle CSV writing with a lock
            async with file_lock:
                if not exists('files/all_option_data.csv'):
                    df.to_csv('files/all_option_data.csv', index=False)
                else:
                    df.to_csv('files/all_option_data.csv', mode='a', header=False, index=False)

        except UniqueViolationError:
            print(f'Skipping {ticker} - Already exists.')
        except Exception as e:
            print(f'Error with {ticker}: {e}')

async def run_all_options():
    await opts.connect()

    
    try:
        ticker_list = []
        tasks = [get_all_options(ticker, ticker_list) for ticker in most_active_tickers]
        await asyncio.gather(*tasks)
    except Exception as e:
        print(f'Error - {e}')

asyncio.run(run_all_options())


from polygon.websocket import WebSocketMessage,WebSocketClient, EquityAgg,EquityQuote,EquityTrade, Market
import pytz

from discord_webhook import AsyncDiscordWebhook, DiscordEmbed
import asyncio


from apis.polygonio.mapping import OPTIONS_EXCHANGES, option_condition_desc_dict, option_condition_dict
from datetime import datetime
from math import isnan
from asyncio import Queue
from datetime import timezone
from polygon.websocket import CryptoTrade
from apis.helpers import calculate_price_to_strike, calculate_days_to_expiry, format_large_number, format_large_numbers_in_dataframe,format_large_numbers_in_dict, get_human_readable_string
import requests
import asyncpg
batch_data_aggs = []
batch_data_trades = []


from pytz import timezone
from apis.helpers import convert_to_ns_datetime
utc = timezone('UTC')
aware_datetime = utc.localize(datetime.utcnow())
from .list_sets import crypto_conditions_dict, crypto_exchanges

async def handle_crypto_msg(m: CryptoTrade):
    conditions = [crypto_conditions_dict.get(i) for i in m.conditions]
    data = { 
        'type': m.event_type,
        'symbol': m.pair,
        'exchange': crypto_exchanges.get(m.exchange),
        'id': m.id,
        'price': m.price,
        'size': m.size,
        'conditions': conditions[0],
        'timestamp': convert_to_ns_datetime(m.timestamp)
    }

    yield data
    # await db_manager.save_structured_message(data, 'crypto_trades')
    # # Send to Flask API

    

    #await data_queue.put(data)




from apis.webull.webull_trading import WebullTrading

import asyncio

wb = WebullTrading()



async def main():
    ticker='AAPL'
    analyst_ratings = await wb.get_analyst_ratings(ticker)


    #DOT NOTATION
    strong_buy = analyst_ratings.strongbuy
    buy = analyst_ratings.buy
    sell = analyst_ratings.sell
    hold = analyst_ratings.hold
    underperform = analyst_ratings.underperform
    print(f"Strong: {strong_buy} | Buy: {buy} | Hold: {hold} | Underperform: {underperform} | Sell: {sell}")


    data_dict = analyst_ratings.data_dict

    df = analyst_ratings.df
    df['ticker'] = ticker

    print(df)




import asyncio

async def all_main():
    # Initialize the WebullTrading class
    webull_trading = WebullTrading()

    # Define a sample ticker symbol for demonstration
    sample_ticker = 'AAPL'

    # Run each function one by one
    bars = await webull_trading.get_bars(sample_ticker)
    print("Bars:", bars)

    stock_quote = await webull_trading.get_stock_quote(sample_ticker)
    print("Stock Quote:", stock_quote.df)

    analyst_ratings = await webull_trading.get_analyst_ratings(sample_ticker)
    print("Analyst Ratings:", analyst_ratings.df)

    short_interest = await webull_trading.get_short_interest(sample_ticker)
    print("Short Interest:", short_interest.df)

    inst_holding = await webull_trading.institutional_holding(sample_ticker)
    print("Institutional Holding:", inst_holding.to_dict())

    vol_analysis = await webull_trading.volume_analysis(sample_ticker)
    print("Volume Analysis:", vol_analysis.df)

    cost_dist = await webull_trading.cost_distribution(sample_ticker)
    print("Cost Distribution:", cost_dist.df)

    news_data = await webull_trading.news(sample_ticker)
    print("News:", news_data.df)

    balance_sheet_data = await webull_trading.balance_sheet(sample_ticker)
    print("Balance Sheet:", balance_sheet_data.df)

    cash_flow_data = await webull_trading.cash_flow(sample_ticker)
    print("Cash Flow:", cash_flow_data.df)

    income_statement_data = await webull_trading.income_statement(sample_ticker)
    print("Income Statement:", income_statement_data.df)

    capital_flow_data = await webull_trading.capital_flow(sample_ticker)
    print("Capital Flow:", capital_flow_data.df)

    etf_holdings_data = await webull_trading.etf_holdings(sample_ticker)
    print("ETF Holdings:", etf_holdings_data.df)




asyncio.run(all_main())
#asyncio.run(main())


import aiohttp
from asyncpg import create_pool
from asyncpg.exceptions import UndefinedTableError
from .ofr_list_sets import NYPD_OFR,FNYR_OFR,MMF_OFR,REPO_OFR,TYLD_OFR
import pandas as pd
import os



db_config = {
    "host": os.environ.get('DB_HOST'), # Default to this IP if 'DB_HOST' not found in environment variables
    "port": int(os.environ.get('DB_PORT')), # Default to 5432 if 'DB_PORT' not found
    "user": os.environ.get('DB_USER'), # Default to 'postgres' if 'DB_USER' not found
    "password": os.environ.get('DB_PASSWORD'), # Use the password from environment variable or default
    "database": os.environ.get('DB_NAME') # Database name for the new jawless database
}



class OFR:
    def __init__(self, connection_string= None):
        self.base_url = "https://data.financialresearch.gov/v1"


        self.endpoints = { 
            'mnemonics': '/metadata/mnemonics',
            'metadata_query': '/metadata/query',
            'metadata_search': '/metadata/search',
            'series_data': '/series/timeseries',
            'series_spread': '/calc/spread',
            'full_series': '/series/full',
            'multifull_series': '/series/multifull',
            'dataset': '/series/dataset'
        }
        self.conn = None
        self.pool = None
        if connection_string is not None:
            self.connection_string = os.environ.get('CONNECTION_STRING')

        self.chat_memory = []  # In-memory list to store chat messages
        self.offset = 0


    async def connect(self):
        # Assuming create_pool is defined somewhere and db_config is provided
        self.pool = await create_pool(
            **db_config, min_size=1, max_size=10
        )
        return self.pool


    async def construct_series_url(
        self,
        mnemonic,
        label="aggregation",
        start_date=None,
        end_date=None,
        periodicity=None,
        how=None,
        remove_nulls=None,
        time_format=None,
    ):
        base_url = "https://data.financialresearch.gov/v1/series/timeseries"
        
        # Initialize query parameters
        params = {}
        
        # Required parameter
        params['mnemonic'] = mnemonic
        
        # Optional parameters
        if label:
            params['label'] = label
        if start_date:
            params['start_date'] = start_date
        if end_date:
            params['end_date'] = end_date
        if periodicity:
            params['periodicity'] = periodicity
        if how:
            params['how'] = how
        if remove_nulls:
            params['remove_nulls'] = remove_nulls
        if time_format:
            params['time_format'] = time_format
            
        # Construct the full URL
        query_string = "&".join(f"{key}={value}" for key, value in params.items())
        full_url = f"{base_url}?{query_string}"
        
        return full_url
    # Function to generate URLs for all mnemonics in a given dictionary
    async def generate_urls_for_mnemonics(self, mnemonic_dict, **kwargs):
        urls = {}
        for mnemonic, series_name in mnemonic_dict.items():
            url = self.construct_series_url(mnemonic, **kwargs)
            urls[series_name] = url
        return urls


    async def query_ofr_api(self, endpoint):
        print(self.base_url+endpoint)
        async with aiohttp.ClientSession() as session:
            async with session.get(self.base_url+endpoint) as resp:
                mnemonics_list = await resp.json()

                return mnemonics_list
    async def insert_ofr_data(self, df: pd.DataFrame):
        async with self.pool.acquire() as conn:
            df['date'] = pd.to_datetime(df['date'])
            df['value'] = df['value'].astype(float)
            try:
                await conn.executemany(
                    """
                    INSERT INTO ofr_data (
                        date, 
                        value, 
                        name, 
                        mnemonic
                    ) VALUES ($1, $2, $3, $4)
                    ON CONFLICT (date, mnemonic) DO UPDATE
                    SET value = EXCLUDED.value, name = EXCLUDED.name;
                    """,
                    list(df.itertuples(index=False, name=None))
                )
            except UndefinedTableError:
                print(f"No table has been created! Creating the table now..")

                # Create the table
                await conn.execute(
                    """
                    CREATE TABLE IF NOT EXISTS ofr_data (
                        date DATE NOT NULL,
                        value FLOAT NOT NULL,
                        name TEXT NOT NULL,
                        mnemonic TEXT NOT NULL,
                        PRIMARY KEY (date, mnemonic)
                    );
                    """
                )

                # Try inserting again
                await conn.executemany(
                    """
                    INSERT INTO ofr_data (
                        date, 
                        value, 
                        name, 
                        mnemonic
                    ) VALUES ($1, $2, $3, $4)
                    ON CONFLICT (date, mnemonic) DO UPDATE
                    SET value = EXCLUDED.value, name = EXCLUDED.name;
                    """,
                    list(df.itertuples(index=False, name=None))
                )

                print(f'Table created successfully.')


import os
from dotenv import load_dotenv
load_dotenv()
from .models.application_commands import ApplicationCommands,Applications
from .models.search import Attachments, Messages
from . import DiscordDBManager
import time
import psycopg2
from .models.webhooks import Webhooks
import random
import requests
from datetime import datetime
import json
class DiscordSDK(DiscordDBManager):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        """
        A toolkit for working with the discord API.

        Your token is retrieved from the developer tools window. 

        >>> Network tab:

        Type a message in a channel while the dev. tools window is open with the 
        "Network" tab selected and filtered by "FETCH/XHR". 


        >>> URL:

        After typing the message - you'll see a URL pop up. Click on it - and scroll 
        down until you see the "Request Headers" section.
        
        Look for "authorization" and copy the token beside it, and paste it into an .env file as YOUR_DISCORD_HTTP_TOKEN=.
        """

    def get_headers(self, api_key):
        """YOU WILL NEED YOUR DISCORD AUTHORIZATION TOKEN FROM THE DEVELOPER TOOLS"""

        
        
        headers = {
                "authority": "discord.com",
                "method": "POST",
                "path": "/api/v9/science",
                "scheme": "https",
                "Accept": "*/*",
                'Content-Type': 'application/json',
                "Accept-Encoding": "gzip, deflate, br",
                "Accept-Language": "en-US,en;q=0.9",
                "Authorization": api_key,
                "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/120.0.0.0 Safari/537.36",

        }

        return headers
    def connect_to_database(self, user:str='chuck', port:int=5432, password:str='fud', host:str='localhost', database:str='markets'):
        """
        Arguments:

        >>> db_config:

         **PASS IN YOUR POSTGRESQL CONFIG INFO:

         >>> database: the database name
         >>> host: the host (localhost)
         >>> port: the port (5432)
         >>> user: your user
         >>> password: your db password
        """
        # Define database connection parameters
        database = "markets"
        user = "chuck"
        port = 5432
        password = "fud"
        host = "localhost"

        # Create a database connection using psycopg2
        try:
            connection = psycopg2.connect(
                database=database,
                user=user,
                port=port,
                password=password,
                host=host
            )
            return connection
        except psycopg2.Error as e:
            print("Error connecting to the database:", e)
            return None
    def get_roles(self, guild_id, role_id):
        """Gets role information for a guild. Use role_counts to get the role IDs."""
        url=f"https://discord.com/api/v9/guilds/{guild_id}/roles/{role_id}"
        r = requests.patch(url, headers=self.headers).json()

        return r
    def search(self, guild_id, author_id:str='375862240601047070', has:str='file'):
        """Searches discord message history
        
        >>> author_id: the id of the author to search

        >>> has: file, embed, link, mentions, image..
        """
        offset = 0
        while True:
            time.sleep(5)
            offset = offset + 25
            url=f"https://discord.com/api/v9/guilds/{guild_id}/messages/search?author_id={author_id}&has={has}&offset={offset}"
            r = requests.get(url, headers=self.headers).json()


            messages = Messages(r)


            attachments = messages.attachments
            attachments = [item for sublist in attachments for item in sublist]

            attachments = Attachments(attachments)

            # Assuming attachments.filename is your list of filenames
            ogg_filenames = [file for file in attachments.filename if file.endswith('.ogg')]
        
            # Now, create or modify the DataFrame
            # If attachments.as_dataframe() provides the full DataFrame, filter it using the ogg_filenames list
            if hasattr(attachments, 'as_dataframe'):
                full_df = attachments.as_dataframe
                # Filter the DataFrame to include only rows with .ogg filenames
                ogg_df = full_df[full_df['filename'].isin(ogg_filenames)]

                for i,row in ogg_df.iterrows():
                    url = row['proxy_url']

                    yield url

            


    def sanitize_filename(self, filename):
        return "".join([c for c in filename if c.isalpha() or c.isdigit() or c in (' ', '.', '_', '-')]).rstrip()

    def get_unique_filename(self, set1, set2, directory):
        max_attempts = 10
        for _ in range(max_attempts):
            current_time = datetime.datetime.now().strftime("%Y%m%d_%H%M%S")
            random_part = f"{random.choice(set1)}f{random.choice(set2)}"
            filename = f"{current_time}_{random_part}.ogg"
            if not os.path.exists(os.path.join(directory, self.sanitize_filename(filename))):
                return filename
        raise Exception("Failed to generate a unique filename after several attempts")



    def download_voice_messages(self):
        set1 = ['A', 'R', 'G', 'B', 'C', 'Q', 'S', 'T', 'B', 'V']
        set2 = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]

        for url in self.search():
            print(f"Processing URL: {url}")

            try:
                response = requests.get(url)

                if response.status_code == 200:
                    directory_path = f"messages/{datetime.datetime.now().strftime('%Y%m%d')}"
                    os.makedirs(directory_path, exist_ok=True)

                    filename = self.get_unique_filename(set1, set2, directory_path)
                    file_path = os.path.join(directory_path, self.sanitize_filename(filename))

                    with open(file_path, 'wb') as file:
                        file.write(response.content)
                    print(f"Downloaded: {file_path}")
                else:
                    print(f"Failed to download from {url}: Status code {response.status_code}")

            except requests.exceptions.RequestException as e:
                print(f"Error downloading from {url}: {e}")
            except Exception as e:
                print(f"Error: {e}")

    def role_counts(self, guild_id):
        url=f"https://discord.com/api/v9/guilds/{guild_id}/roles/member-counts"
        r = requests.get(url, headers=self.headers).json()
        for i in r:
            print(i)

    def create_channel(self, guild_id, name, type='4', channel_description:str=None, with_webhook:bool=False, webhook_name:str=None, permission_overwrites:str=None):
        headers = self.get_headers(os.environ.get('DISCORD_AUTHORIZATION'))
        """
        Create a channel in a Discord guild.
        
        Args:
            >>> guild_id (str): The ID of the guild where the channel will be created.
            >>> name (str): The name of the channel.
            >>> type (str): The type of channel ('0' for text, '4' for category, '2' for voice, ''). Defaults to '4'.
        
        Returns:
            Response from the Discord API.
        """



        # Prepare the payload
        payload = {
            "type": type,
            "name": name,
            "description": channel_description,

        }
        # Generate permission_overwrites from self.role_dict

        if permission_overwrites is not None:
            permission_overwrites = [{
                "id": role_id,
                "type": 0,
                "deny": "0",
                "allow": "71453223935041"
            } for role_id in self.role_dict.values()]

            payload['permission_overwrites'] = permission_overwrites

        # Make the API request to create the channel
        r = requests.post(f"https://discord.com/api/v9/guilds/{guild_id}/channels", headers=headers, data=json.dumps(payload))

        id = r.json()['id']


        if with_webhook != False and name is not None:
            self.create_webhook(channel_id=id, name=webhook_name)


            return "Webhook created successfully!"

        


    def application_commands(self, guild):
        """
        >>> Returns: applications_df and commands_df

        >>> usage: applications, commands = application_commands(guild)
        
        """
        url=f"https://discord.com/api/v9/guilds/{guild}/application-command-index"
        r = requests.get(url, headers=self.headers).json()
        applications = Applications(r['applications'])
        application_commands = r['application_commands']
        commands = ApplicationCommands(application_commands)

        applications_df = applications.applications_as_dataframe
        commands_df = commands.as_dataframe



        return applications_df, commands_df


    def get_webhooks(self, guild_id):
        url=f"https://discord.com/api/v9/guilds/{guild_id}/webhooks"
        r = requests.get(url, headers=self.headers).json()


        webhooks = Webhooks(r)


        return webhooks.as_dataframe
    


    def create_webhook(self, channel_id, name):
        """
        Create a webhook in a Discord channel.

        This function sends a POST request to the Discord API to create a webhook
        in the specified channel with the provided name.

        :param channel_id: The ID of the Discord channel where the webhook will be created.
        :param name: The name of the webhook to be created.
        :return: A JSON response containing information about the created webhook.
        """
        url = f"https://discord.com/api/v9/channels/{channel_id}/webhooks"
        payload = {'name': name}
        r = requests.post(url, headers=self.headers, data=json.dumps(payload))

        return r.json()


    def delete_channel(self, channel_id):
        
        url = f"https://discord.com/api/v9/channels/{channel_id}"


        r = requests.delete(url=url, headers=self.headers)


        return r
    
    def create_thread(self, channel_id:str="1193718649404538980", name:str="YOUR THREAD NAME HERE", topic:str='YOUR DESCRIPTION HERE', nsfw:bool=False, invitable:bool=True, locked:bool=False):
        connection = self.connect_to_database()
        self.create_thread_table(connection=connection)
        payload = {"name":name,"type":11,"topic":topic,"bitrate":64000,"user_limit":0,"nsfw":nsfw,"flags":0,"rate_limit_per_user":0,"auto_archive_duration":10080,"locked": locked,"invitable":invitable}
        r = requests.post(f"https://discord.com/api/v9/channels/{channel_id}/threads", headers=self.headers, data=json.dumps(payload)).json()

        thread_id = r.get('id')
        parent_channel = r.get('parent_id')
        name = r.get('name')
        total_messages = r.get('total_messages_sent')
        message_count = r.get('message_count')
        last_message_id = r.get('last_message_id')
        guild = r.get('guild_id')

        # Insert the retrieved attributes into the database
        self.insert_thread(thread_id, parent_channel, name, total_messages, message_count, last_message_id, guild)


        

    def upload_file(self, channel_id, file_path):
        random_id = random.randint(1, 9223372036854775807)
        file_size = os.path.getsize(file_path)
        url=f"https://discord.com/api/v9/channels/{channel_id}/attachments"

        payload = {"files":[{"filename":file_path,"file_size":file_size,"id":random_id,"is_clip":False}]}
        
        r = requests.post(url, headers=self.headers, data=json.dumps(payload)).json()
        
        attachments = r['attachments']

        upload_url = [i.get('upload_url') for i in attachments]

        return upload_url
    



import sys
import os

sys.path.append(os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__)))))
from asyncpg import create_pool
import requests
import matplotlib.pyplot as plt
from decimal import Decimal
import matplotlib.dates as mdates
import aiohttp
import asyncio
from .treasury_models import AvgInterestRates
from asyncpg.exceptions import UndefinedTableError
from datetime import datetime, timedelta
import pandas as pd
from datetime import datetime

from database_manager import DatabaseManager



db_config = {
    "host": os.environ.get('DB_HOST'), # Default to this IP if 'DB_HOST' not found in environment variables
    "port": int(os.environ.get('DB_PORT')), # Default to 5432 if 'DB_PORT' not found
    "user": os.environ.get('DB_USER'),
    "password": os.environ.get('DB_PASSWORD'), # Use the password from environment variable or default
    "database": os.environ.get('DB_NAME') # Database name for the new jawless database
}

db_manager = DatabaseManager(**db_config)


# Determine the table name based on the endpoint
endpoint_to_table = {
    "/v1/accounting/od/securities_sales": "Sales",
    "/v1/accounting/od/securities_sales_term": "Sales by Term",
    "/v1/accounting/od/securities_transfers": "Transfers of Marketable Securities",
    "/v1/accounting/od/securities_conversions": "Conversions of Paper Savings Bonds",
    "/v1/accounting/od/securities_redemptions": "Redemptions",
    "/v1/accounting/od/securities_outstanding": "Outstanding",
    "/v1/accounting/od/securities_c_of_i": "Certificates of Indebtedness",
    "/v1/accounting/od/securities_accounts": "Accounts",
}

class Treasury:
    def __init__(self, host:str=os.environ.get('DB_HOST'), port:str=os.environ.get('DB_PORT'), user:str=os.environ.get('DB_USER'), password:str=os.environ.get('DB_USER'), database:str=os.environ.get('DB_MNAME')):
        self.today = datetime.now().strftime('%Y-%m-%d')
        self.yesterday = (datetime.now() - timedelta(days=1)).strftime('%Y-%m-%d')
        self.tomorrow = (datetime.now() + timedelta(days=1)).strftime('%Y-%m-%d')
        self.thirty_days_ago = (datetime.now() - timedelta(days=30)).strftime('%Y-%m-%d')
        self.thirty_days_from_now = (datetime.now() + timedelta(days=30)).strftime('%Y-%m-%d')
        self.fifteen_days_ago = (datetime.now() - timedelta(days=15)).strftime('%Y-%m-%d')
        self.fifteen_days_from_now = (datetime.now() + timedelta(days=15)).strftime('%Y-%m-%d')
        self.eight_days_from_now = (datetime.now() + timedelta(days=8)).strftime('%Y-%m-%d')
        self.eight_days_ago = (datetime.now() - timedelta(days=8)).strftime('%Y-%m-%d')


        self.base_url = "https://api.fiscaldata.treasury.gov/services/api/fiscal_service/"
        self.conn = None
        self.pool = None
        self.host = host
        self.port = port
        self.user = user
        self.password = password
        self.database = database
        self.connection_string = f"postgresql://{user}:{password}@{host}:{port}/{database}"
        
        self.chat_memory = []  # In-memory list to store chat messages


        self.endpoint_to_table = {
            "/v1/accounting/od/securities_sales": "td_sales",
            "/v1/accounting/od/securities_sales_term": "td_sales_term",
            "/v1/accounting/od/securities_transfers": "td_transfers",
            "/v1/accounting/od/securities_conversions": "td_conversions",
            "/v1/accounting/od/securities_redemptions": "td_redemptions",
            "/v1/accounting/od/securities_outstanding": "td_outstanding",
            "/v1/accounting/od/securities_c_of_i": "td_indebtedness",
            "/v1/accounting/od/securities_accounts": "td_accounts",
            '/v2/accounting/od/gold_reserve': "fed_gold_vault",
            '/v1/accounting/tb/pdo2_offerings_marketable_securities_other_regular_weekly_treasury_bills': "marketable_securities",
        }



    async def connect(self):
        self.pool = await create_pool(
            **db_config, min_size=1, max_size=10
        )

        return self.pool

    async def disconnect(self):
        await self.pool.close()
    async def insert_treasury_data(self, table_name, data):
        try:
            # Convert 'record_date' from string to date object if it exists in data
            if 'record_date' in data:
                data['record_date'] = datetime.strptime(data['record_date'], '%Y-%m-%d').date()

            # Corrected the keys for 'auction_date' and 'issue_date'
            if 'auction_date' in data:
                data['auction_date'] = datetime.strptime(data['auction_date'], '%Y-%m-%d').date()

            if 'issue_date' in data:
                data['issue_date'] = datetime.strptime(data['issue_date'], '%Y-%m-%d').date()

            # Convert known fields that should be integers
            # Convert known fields that should be integers
            integer_fields = [
                'securities_sold_cnt', 'src_line_nbr', 'record_fiscal_year', 'record_fiscal_quarter',
                'record_calendar_year', 'record_calendar_quarter', 'record_calendar_month', 'record_calendar_day',
                'securities_redeemed_cnt', 'from_legacy_system_cnt', 'from_commercial_book_entry_cnt',
                'total_incoming_transfers_cnt', 'unrestrict_primary_accts_cnt', 'funded_accts_cnt','security_term_nbr','paper_sb_conversions_cnt'
            ]

            float_fields = ['securities_sold_amt', 'securities_redeemed_amt']

            decimal_fields = [
                'outstanding_cert_indebt_amt',  # Add other decimal fields if needed
            ]

            for field in integer_fields:
                if field in data:
                    data[field] = int(data[field]) if data[field] != 'null' else None

            for field in float_fields:
                if field in data:
                    data[field] = float(data[field]) if data[field] != 'null' else None

            for field in decimal_fields:
                if field in data:
                    data[field] = Decimal(data[field]) if data[field] != 'null' else None
            async with self.pool.acquire() as conn:
                # Construct the insert query dynamically based on the data dictionary
                columns = ', '.join(data.keys())
                placeholders = ', '.join([f"${i+1}" for i in range(len(data))])

                insert_query = f"INSERT INTO {table_name} ({columns}) VALUES ({placeholders}) ON CONFLICT DO NOTHING;"

                print(f"Debug - SQL Query: {insert_query}")
                print(f"Debug - Data Values: {list(data.values())}")

                await conn.execute(insert_query, *data.values())

            print(f"Data inserted into {table_name} successfully.")

        except Exception as e:
            print(f"Error inserting data into {table_name}: {e}")
            print(f"ATTEMPTING to create table...")
            await db_manager.create_table(data, table_name='fed_gold', unique_column='insertion_timestamp')

    async def fetch_gold_data(self):
        await self.connect()
        async with self.pool.acquire() as conn:
            query = "SELECT record_date, book_value_amt, fine_troy_ounce_qty FROM fed_gold ORDER BY record_date ASC;"
            try:
                rows = await conn.fetch(query)

                
                return rows
            except UndefinedTableError:
                print(f'Tables doesnt exist!')
          
    async def plot_gold_data(self):
        # Step 1: Fetch data
        data = await self.fetch_gold_data()

        # Step 2: Convert to DataFrame
        df = pd.DataFrame(data, columns=['record_date', 'book_value_amt', 'fine_troy_ounce_qty'])
        
        # Convert 'record_date' to datetime format for better x-axis formatting
        df = df.sort_values(by='record_date')

        # Plot
        fig, ax1 = plt.subplots(figsize=(14, 8))

        # Plotting fine_troy_ounce_qty
        ax1.plot(df['record_date'], df['fine_troy_ounce_qty'], 'b-', marker='o', label='Fine Troy Ounce Quantity')
        ax1.set_xlabel('Record Date')
        ax1.set_ylabel('Fine Troy Ounce Quantity', color='b')
        ax1.tick_params('y', colors='b')



        # Title and grid
        plt.title('Fine Troy Ounce Quantity vs Book Value Amount Over Time')
        ax1.grid(True)

        # Legend
        ax1.legend(loc='upper left', bbox_to_anchor=(0.0,1), fontsize=10)


        plt.show()


        
    def data_act_compliance(self):
        url=self.base_url+f"/v2/debt/tror/data_act_compliance?filter=record_date:gte:2018-07-01,record_date:lte:{self.today}&sort=-record_date,agency_nm,agency_bureau_indicator,bureau_nm"
        r = requests.get(url).json()
        data = r['data']
        df = pd.DataFrame(data)

        return df



    async def query_treasury(self, endpoint):
        await self.connect()
        

        url = self.base_url + endpoint + f"?sort=-record_date" #f"?filter=record_date:gte:{two_years_ago_str}"
        
        print(url)
        
        table_name = self.endpoint_to_table.get(endpoint, None)
        
        async with aiohttp.ClientSession() as session:
            try:
                async with session.get(url) as resp:
                    r = await resp.json()
                    data = r['data']
                    print(data)

                    if table_name:
                        for record in data:
                            await self.insert_treasury_data(table_name, record)
                    else:
                        print(f"Table name not found for endpoint: {endpoint}")

            except aiohttp.ClientError as e:
                print(f"Error fetching data from {url}: {e}")
            except Exception as e:
                print(f"Error: {e}")



    def avg_interest_rates(self):
        """Gets avg. interest rates for US treasury"""

        r = requests.get("https://api.fiscaldata.treasury.gov/services/api/fiscal_service/v2/accounting/od/avg_interest_rates?sort=-record_date").json()
        data = r['data']


        return AvgInterestRates(data)





from apis.ofr.ofr_sdk import OFR
from apis.ofr.ofr_list_sets import REPO_OFR,NYPD_OFR,TYLD_OFR,MMF_OFR,FNYR_OFR

import asyncio
import aiohttp

import pandas as pd



sdk = OFR()
connection_string = None
async def main():
    await sdk.connect()
    mnemonics = await sdk.query_ofr_api(sdk.endpoints.get('mnemonics'))
    for mnemonic in mnemonics:
        series_data = await sdk.construct_series_url(mnemonic)
        async with aiohttp.ClientSession() as session:
            async with session.get(series_data) as resp:
                if str(mnemonic).startswith('REPO'):


                    name = REPO_OFR[mnemonic]
                    print(name)

                    data = await resp.json()
                    print(data)
                    df = pd.DataFrame(data, columns=['date', 'value'])

                    df['name'] = name
                    df['mnemonic'] = mnemonic
                    df['value'] = df['value'].apply(lambda x: format(float(x), '.2f') if x is not None else x)
                    df = df[::-1]
                    print(df)
                    if connection_string is not None:
                        await sdk.insert_ofr_data(df)
                elif str(mnemonic).startswith('TYLD'):
                    name = TYLD_OFR[mnemonic]
                    print(name)
                    data = await resp.json()
                    print(data)
                    df = pd.DataFrame(data, columns=['date', 'value'])

                    df['name'] = name
                    df['mnemonic'] = mnemonic
                    df['value'] = df['value'].apply(lambda x: format(float(x), '.2f') if x is not None else x)
                    df = df[::-1]
                    print(df)
                    if connection_string is not None:
                        await sdk.insert_ofr_data(df)
                elif str(mnemonic).startswith('MMF'):
                    name = MMF_OFR[mnemonic]
                    print(name)
                    data = await resp.json()
                    print(data)
                    df = pd.DataFrame(data, columns=['date', 'value'])

                    df['name'] = name
                    df['mnemonic'] = mnemonic
                    df['value'] = df['value'].apply(lambda x: format(float(x), '.2f') if x is not None else x)
                    df = df[::-1]
                    print(df)
                    if connection_string is not None:
                        await sdk.insert_ofr_data(df)
                elif str(mnemonic).startswith('FNYR'):
                    name = FNYR_OFR[mnemonic]
                    print(name)
                    data = await resp.json()
                    print(data)

                    df = pd.DataFrame(data, columns=['date', 'value'])

                    df['name'] = name
                    df['mnemonic'] = mnemonic
                    df['value'] = df['value'].apply(lambda x: format(float(x), '.2f') if x is not None else x)
                    df = df[::-1]
                    print(df)
                    if connection_string is not None:
                        await sdk.insert_ofr_data(df)


                elif str(mnemonic).startswith('NYPD'):
                    name = NYPD_OFR[mnemonic]
                    print(name)
                    data = await resp.json()
                    print(data)

                    df = pd.DataFrame(data, columns=['date', 'value'])

                    df['name'] = name
                    df['mnemonic'] = mnemonic
                    df['value'] = df['value'].apply(lambda x: format(float(x), '.2f') if x is not None else x)
                    df = df[::-1]
                    print(df)
                    if connection_string is not None:
                        await sdk.insert_ofr_data(df)


asyncio.run(main())


from dataclasses import dataclass, asdict
from typing import Optional

@dataclass
class CombinedCompanyResults:
    ticker: Optional[str] = None
    name: Optional[str] = None
    market: Optional[str] = None
    locale: Optional[str] = None
    primary_exchange: Optional[str] = None
    type: Optional[str] = None
    active: Optional[bool] = None
    currency_name: Optional[str] = None
    cik: Optional[str] = None
    composite_figi: Optional[str] = None
    share_class_figi: Optional[str] = None
    market_cap: Optional[int] = None
    phone_number: Optional[str] = None
    description: Optional[str] = None
    sic_code: Optional[str] = None
    sic_description: Optional[str] = None
    ticker_root: Optional[str] = None
    homepage_url: Optional[str] = None
    total_employees: Optional[int] = None
    list_date: Optional[str] = None
    share_class_shares_outstanding: Optional[int] = None
    weighted_shares_outstanding: Optional[int] = None
    round_lot: Optional[int] = None
    
    # Address attributes
    address1: Optional[str] = None
    city: Optional[str] = None
    state: Optional[str] = None
    postal_code: Optional[str] = None
    
    # Branding attributes
    logo_url: Optional[str] = None
    icon_url: Optional[str] = None
    
    def to_dict(self):
        return asdict(self)


import os
from dotenv import load_dotenv
load_dotenv()
from fudstop.apis.polygonio.async_polygon_sdk import Polygon
from fudstop.list_sets.ticker_lists import most_active_tickers
polygon = Polygon(connection_string=os.environ.get('POLYGON_STRING'))
import aiohttp
import json
import asyncio

async def send_discord_webhook(webhook_url, message, color):
    async with aiohttp.ClientSession() as session:
        webhook_data = {
            "embeds": [{
                "title": "RSI Alert",
                "description": message,
                "color": color
            }]
        }
        async with session.post(webhook_url, json=webhook_data) as response:
            return await response.text()

async def process_ticker(i, timespan, webhook_url):
    rsi = i.rsi_value[0]
    status = ''
    if rsi <= 30 and timespan == 'hour':
        status = 'is oversold'
    elif rsi >= 70 and timespan == 'hour':
        status = 'is overbought'

    if status:
        color = 0x00ff00 if status == 'is oversold' else 0xff0000
        message = f"{i.ticker} {status} on the {timespan} with an RSI of: {rsi}"
        await send_discord_webhook(webhook_url, message, color)

async def gather_rsi_results(timespan, webhook_url):
    timespans = ['minute', 'day', 'week', 'hour']
    rsi = await polygon.gather_rsi_for_all_tickers(tickers=most_active_tickers, timespans=timespans)
    tasks = []
    for i in rsi:
        if hasattr(i, 'rsi_value'):
            if i and i.rsi_value and len(i.rsi_value) > 0:
                task = asyncio.create_task(process_ticker(i, timespan, webhook_url))
                tasks.append(task)
    await asyncio.gather(*tasks)

async def main():
    hour_hook_url = os.environ.get('osob_hour')
    minute_hook_url = os.environ.get('osob_minute')
    day_hook_url = os.environ.get('osob_day')
    week_hook_url = os.environ.get('osob_week')
    
    await asyncio.gather(
        gather_rsi_results('hour', hour_hook_url),
        gather_rsi_results('minute', minute_hook_url),
        gather_rsi_results('day', day_hook_url)
    )

asyncio.run(main())


from .accounting import AccountingSelect
from .gex import Gex, GexSelect
from .yfin import yfCOG, YFSelect
from .earnings import EarningsSelect



import disnake
from disnake import TextInputStyle
from disnake.ext import commands


class MainView(disnake.ui.View):
    def __init__(self, ticker=None):
        self.ticker=ticker
        super().__init__(timeout=None)


    @disnake.ui.button(style=disnake.ButtonStyle.blurple, label='-', custom_id='-1', row=0, disabled=True)
    async def one(self, button:disnake.ui.Button, inter:disnake.AppCmdInter):
        pass

    @disnake.ui.button(style=disnake.ButtonStyle.grey, label='-', custom_id='-2', row=0, disabled=True)
    async def two(self, button:disnake.ui.Button, inter:disnake.AppCmdInter):
        pass

    @disnake.ui.button(style=disnake.ButtonStyle.blurple, label='-', custom_id='-3', row=0, disabled=True)
    async def three(self, button:disnake.ui.Button, inter:disnake.AppCmdInter):
        pass

    @disnake.ui.button(style=disnake.ButtonStyle.grey, label='-', custom_id='-4', row=0, disabled=True)
    async def four(self, button:disnake.ui.Button, inter:disnake.AppCmdInter):
        pass

    @disnake.ui.button(style=disnake.ButtonStyle.blurple, label='-', custom_id='-5', row=0, disabled=True)
    async def five(self, button:disnake.ui.Button, inter:disnake.AppCmdInter):
        pass

    @disnake.ui.button(style=disnake.ButtonStyle.green, label='GEX', custom_id='gex', row=1)
    async def gex(self, button:disnake.ui.Button, inter:disnake.AppCmdInter):
        await inter.response.edit_message(view=GexView(self.ticker))


    @disnake.ui.button(style=disnake.ButtonStyle.green, label='INFO', custom_id='info', row=1)
    async def info(self, button:disnake.ui.Button, inter:disnake.AppCmdInter):
        await inter.response.edit_message(view=YFView(self.ticker))

    @disnake.ui.button(style=disnake.ButtonStyle.green, label='REG', custom_id='reg', row=1)
    async def reg(self, button:disnake.ui.Button, inter:disnake.AppCmdInter):
        pass


    @disnake.ui.button(style=disnake.ButtonStyle.green, label='FS', custom_id='fiscal', row=1)
    async def fs(self, button:disnake.ui.Button, inter:disnake.AppCmdInter):
        pass

    @disnake.ui.button(style=disnake.ButtonStyle.green, label='ACC', custom_id='accounting', row=1)
    async def acc(self, button:disnake.ui.Button, inter:disnake.AppCmdInter):
        await inter.response.edit_message(view=AccountingView())
        


class GexView(disnake.ui.View):
    def __init__(self, ticker=None):
        self.ticker=ticker
        super().__init__(timeout=None)


        self.add_item(GexSelect(self.ticker))


    @disnake.ui.button(style=disnake.ButtonStyle.blurple, emoji='♻️', custom_id='toMainView', row=3)
    async def recycler(self, button: disnake.ui.Button, inter:disnake.AppCmdInter):
        await inter.response.edit_message(view=MainView(self.ticker))




class AccountingView(disnake.ui.View):
    def __init__(self):
     
        super().__init__(timeout=None)


        self.add_item(AccountingSelect())


    @disnake.ui.button(style=disnake.ButtonStyle.blurple, emoji='♻️', custom_id='toMainView', row=3)
    async def recycler(self, button: disnake.ui.Button, inter:disnake.AppCmdInter):
        await inter.response.edit_message(view=MainView())


class EarningsView(disnake.ui.View):
    def __init__(self):
     
        super().__init__(timeout=None)


        self.add_item(EarningsSelect())


    @disnake.ui.button(style=disnake.ButtonStyle.blurple, emoji='♻️', custom_id='toMainView', row=3)
    async def recycler(self, button: disnake.ui.Button, inter:disnake.AppCmdInter):
        await inter.response.edit_message(view=MainView())




class YFView(disnake.ui.View):
    def __init__(self, ticker=None):
        self.ticker=ticker
        super().__init__(timeout=None)

        self.add_item(YFSelect(ticker=self.ticker))

    @disnake.ui.button(style=disnake.ButtonStyle.blurple, emoji='♻️', custom_id='toMainView', row=3)
    async def recycler(self, button: disnake.ui.Button, inter:disnake.AppCmdInter):
        await inter.response.edit_message(view=MainView(self.ticker))









import os

from dotenv import load_dotenv
load_dotenv()
import pandas as pd
from fudstop.apis.polygonio.polygon_options import PolygonOptions
from datetime import datetime
from discord_webhook import AsyncDiscordWebhook
from fudstop.apis.gexbot.gexbot import GEXBot
opts = PolygonOptions()
gex = GEXBot()
import asyncio




# Initialize previous values to None

# Global dictionary to maintain previous values for each ticker
prev_values = {}

async def main(ticker):
    global prev_values  # Reference the global prev_values dictionary

    # Initialize prev_values for the ticker if not already done
    if ticker not in prev_values:
        prev_values[ticker] = {
            "major_pos_vol": None,
            "major_neg_vol": None,
            "major_pos_oi": None,
            "major_neg_oi": None,
            "zero_gamma": None,
        }

    # Get the latest data
    df = await gex.major_levels(ticker)
    df = df.transpose()

    # Extract current values
    current_values = {
        "major_pos_vol": df['major_pos_vol'].iloc[0],
        "major_neg_vol": df['major_neg_vol'].iloc[0],
        "major_pos_oi": df['major_pos_oi'].iloc[0],
        "major_neg_oi": df['major_neg_oi'].iloc[0],
        "zero_gamma": df['zero_gamma'].iloc[0]
    }

    # Check for any changes and build the message
    message_content = f"> GEX CHANGE: \n"
    change_detected = False
    for key in current_values:
        if current_values[key] != prev_values[ticker].get(key):
            change_detected = True
            prev_value = prev_values[ticker].get(key, 'N/A')
            current_value = current_values[key]
            message_content += f"> {key.replace('_', ' ').title()}: Prev: **${round(float(prev_value),2) if prev_value is not None else prev_value}** > Now: **${round(float(current_value),2)}**\n"

    # Send the message to Discord only if there's a change
    if change_detected:
        webhook = AsyncDiscordWebhook(
            "https://discord.com/api/webhooks/1126626151750709368/NHco29Iobl8BbwiQfZ2iTZTi6pi16Lpe1b0_4tVN1f-BbD7MzpI3IlaW50jJzgNJvaqg",
            content=message_content
        )
        await webhook.execute()

        # Update the previous values
        prev_values[ticker] = current_values.copy()

async def run_main():
    ticker_list = ['SPY', 'SPX']

    while True:
        start_time = datetime.now()

        # Create and run tasks for each ticker
        tasks = [main(ticker) for ticker in ticker_list]
        await asyncio.gather(*tasks)

        # Calculate elapsed time and wait if needed
        elapsed_time = (datetime.now() - start_time).total_seconds()
        if elapsed_time < 60:
            await asyncio.sleep(60 - elapsed_time)

asyncio.run(run_main())


import sys
from pathlib import Path

# Add the project directory to the sys.path
project_dir = str(Path(__file__).resolve().parents[1])
if project_dir not in sys.path:
    sys.path.append(project_dir)
from tabulate import tabulate
from dotenv import load_dotenv
load_dotenv()
from .models.technicals import RSI
from aiohttp.client_exceptions import ClientConnectionError, ClientConnectorError, ClientError, ClientOSError,ContentTypeError
import os
from .polygon_helpers import get_human_readable_string
import sys
import pandas as pd
from fudstop.all_helpers import chunk_string
from pathlib import Path
import aiohttp
import asyncpg
import asyncio
from asyncpg.exceptions import UniqueViolationError
from asyncpg import create_pool
from asyncio import Lock
from urllib.parse import urlencode
from datetime import datetime, timedelta
from ..webull.webull_trading import WebullTrading
from ..y_finance.yf_sdk import yfSDK
import numpy as np
from .models.option_models.universal_snapshot import UniversalOptionSnapshot, UniversalSnapshot
from .polygon_helpers import flatten_nested_dict, flatten_dict
lock = Lock()
trading = WebullTrading()
sema = asyncio.Semaphore(4)
# Function to map Pandas dtypes to PostgreSQL types
def dtype_to_postgres(dtype):
    if pd.api.types.is_integer_dtype(dtype):
        return 'INTEGER'
    elif pd.api.types.is_float_dtype(dtype):
        return 'REAL'
    elif pd.api.types.is_bool_dtype(dtype):
        return 'BOOLEAN'
    elif pd.api.types.is_datetime64_any_dtype(dtype):
        return 'TIMESTAMP'
    elif pd.api.types.is_string_dtype(dtype):
        return 'TEXT'
    else:
        return 'TEXT'  # Default type
class PolygonOptions:
    def __init__(self,user:str='chuck', database:str='markets'):
        self.conn = None
        self.pool = None
        self.connection_string = os.environ.get('POLYGON_STRING')



        self.api_key = os.environ.get('YOUR_POLYGON_KEY')
        self.today = datetime.now().strftime('%Y-%m-%d')
        self.yesterday = (datetime.now() - timedelta(days=1)).strftime('%Y-%m-%d')
        self.tomorrow = (datetime.now() + timedelta(days=1)).strftime('%Y-%m-%d')
        self.thirty_days_ago = (datetime.now() - timedelta(days=30)).strftime('%Y-%m-%d')
        self.thirty_days_from_now = (datetime.now() + timedelta(days=30)).strftime('%Y-%m-%d')
        self.fifteen_days_ago = (datetime.now() - timedelta(days=15)).strftime('%Y-%m-%d')
        self.fifteen_days_from_now = (datetime.now() + timedelta(days=15)).strftime('%Y-%m-%d')
        self.eight_days_from_now = (datetime.now() + timedelta(days=8)).strftime('%Y-%m-%d')
        self.eight_days_ago = (datetime.now() - timedelta(days=8)).strftime('%Y-%m-%d')
        self.one_year_from_now = (datetime.now() + timedelta(days=365)).strftime('%Y-%m-%d')
        self.one_year_ago = (datetime.now() - timedelta(days=365)).strftime('%Y-%m-%d')
        self.yf = yfSDK()


        self.db_config = {
            "host": os.environ.get('localhost'), # Default to this IP if 'DB_HOST' not found in environment variables
            "port": int(os.environ.get('DB_PORT')), # Default to 5432 if 'DB_PORT' not found
            "user": os.environ.get(user), # Default to 'postgres' if 'DB_USER' not found
            "password": os.environ.get('fud'), # Use the password from environment variable or default
            "database": os.environ.get(f'{database}') # Database name for the new jawless database
        }


    def sanitize_value(self, value, col_type):
        """Sanitize and format the value for SQL query."""
        if col_type == 'str':
            # For strings, add single quotes
            return f"'{value}'"
        elif col_type == 'date':
            # For dates, format as 'YYYY-MM-DD'
            if isinstance(value, str):
                try:
                    datetime.strptime(value, '%Y-%m-%d')
                    return f"'{value}'"
                except ValueError:
                    raise ValueError(f"Invalid date format: {value}")
            elif isinstance(value, datetime):
                return f"'{value.strftime('%Y-%m-%d')}'"
        else:
            # For other types, use as is
            return str(value)

    async def connect(self):
       
        self.pool = await asyncpg.create_pool(host='localhost', user='chuck', database='markets', port=5432, password='fud')

        return self.pool
    async def update_options(self, ticker):
        async with sema:
            all_options = await self.find_symbols(ticker)
            df = all_options.df
            print(df)
            await self.batch_insert_dataframe(df, table_name='poly_options', unique_columns='option_symbol')

    
    async def filter_options(self,select_columns:str=None, **kwargs):
        """
        Filters the options table based on provided keyword arguments.
        Usage example:
            await filter_options(strike_price_min=100, strike_price_max=200, call_put='call',
                                 expire_date='2023-01-01', delta_min=0.1, delta_max=0.5)
        """
        # Start with the base query
        query = f"SELECT * FROM poly_options WHERE "
        params = []
        param_index = 1

        # Mapping kwargs to database columns and expected types, including range filters
        column_types = {
            'ticker': ('ticker', 'string'),
            'strike': ('strike', 'float'),
            'strike_min': ('strike', 'float'),
            'strike_max': ('strike', 'float'),
            'expiry': ('expiry', 'date'),
            'expiry_min': ('expiry', 'date'),
            'expiry_max': ('expiry', 'date'),
            'open': ('open', 'float'),
            'open_min': ('open', 'float'),
            'open_max': ('open', 'float'),
            'high': ('high', 'float'),
            'high_min': ('high', 'float'),
            'high_max': ('high', 'float'),
            'low': ('low', 'float'),
            'low_min': ('low', 'float'),
            'low_max': ('low', 'float'),
            'oi': ('oi', 'float'),
            'oi_min': ('oi', 'float'),
            'oi_max': ('oi', 'float'),
            'vol': ('vol', 'float'),
            'vol_min': ('vol', 'float'),
            'vol_max': ('vol', 'float'),
            'delta': ('delta', 'float'),
            'delta_min': ('delta', 'float'),
            'delta_max': ('delta', 'float'),
            'vega': ('vega', 'float'),
            'vega_min': ('vega', 'float'),
            'vega_max': ('vega', 'float'),
            'iv': ('iv', 'float'),
            'iv_min': ('iv', 'float'),
            'iv_max': ('iv', 'float'),
            'dte': ('dte', 'string'),
            'dte_min': ('dte', 'string'),
            'dte_max': ('dte', 'string'),
            'gamma': ('gamma', 'float'),
            'gamma_min': ('gamma', 'float'),
            'gamma_max': ('gamma', 'float'),
            'theta': ('theta', 'float'),
            'theta_min': ('theta', 'float'),
            'theta_max': ('theta', 'float'),
            'sensitivity': ('sensitivity', 'float'),
            'sensitivity_max': ('sensitivity', 'float'),
            'sensitivity_min': ('sensitivity', 'float'),
            'bid': ('bid', 'float'),
            'bid_min': ('bid', 'float'),
            'bid_max': ('bid', 'float'),
            'ask': ('ask', 'float'),
            'ask_min': ('ask', 'float'),
            'ask_max': ('ask', 'float'),
            'close': ('close', 'float'),
            'close_min': ('close', 'float'),
            'close_max': ('close', 'float'),
            'cp': ('cp', 'string'),
            'time_value': ('time_value', 'float'),
            'time_value_min': ('time_value', 'float'),
            'time_value_max': ('time_value', 'float'),
            'moneyness': ('moneyness', 'string'),
            'exercise_style': ('exercise_style', 'string'),
            'option_symbol': ('option_symbol', 'string'),
            'theta_decay_rate': ('theta_decay_rate', 'float'),
            'theta_decay_rate_min': ('theta_decay_rate', 'float'),
            'theta_decay_rate_max': ('theta_decay_rate', 'float'),
            'delta_theta_ratio': ('delta_theta_ratio', 'float'),
            'delta_theta_ratio_min': ('delta_theta_ratio', 'float'),
            'delta_theta_ratio_max': ('delta_theta_ratio', 'float'),
            'gamma_risk': ('gamma_risk', 'float'),
            'gamma_risk_min': ('gamma_risk', 'float'),
            'gamma_risk_max': ('gamma_risk', 'float'),
            'vega_impact': ('vega_impact', 'float'),
            'vega_impact_min': ('vega_impact', 'float'),
            'vega_impact_max': ('vega_impact', 'float'),
            'intrinsic_value_min': ('intrinsic_value', 'float'),
            'intrinsic_value_max': ('intrinsic_value', 'float'),
            'intrinsic_value': ('intrinsic_value', 'float'),
            'extrinsic_value': ('extrinsic_value', 'float'),
            'extrinsic_value_min': ('extrinsic_value', 'float'),
            'extrinsic_value_max': ('extrinsic_value', 'float'),
            'leverage_ratio': ('leverage_ratio', 'float'),
            'leverage_ratio_min': ('leverage_ratio', 'float'),
            'leverage_ratio_max': ('leverage_ratio', 'float'),
            'vwap': ('vwap', 'float'),
            'vwap_min': ('vwap', 'float'),
            'vwap_max': ('vwap', 'float'),
            'price': ('price', 'float'),
            'price_min': ('price', 'float'),
            'price_max': ('price', 'float'),
            'trade_size': ('trade_size', 'float'),
            'trade_size_min': ('trade_size', 'float'),
            'trade_size_max': ('trade_size', 'float'),
            'spread': ('spread', 'float'),
            'spread_min': ('spread', 'float'),
            'spread_max': ('spread', 'float'),
            'spread_pct': ('spread_pct', 'float'),
            'spread_pct_min': ('spread_pct', 'float'),
            'spread_pct_max': ('spread_pct', 'float'),
            'bid_size': ('bid_size', 'float'),
            'bid_size_min': ('bid_size', 'float'),
            'bid_size_max': ('bid_size', 'float'),
            'ask_size': ('ask_size', 'float'),
            'ask_size_min': ('ask_size', 'float'),
            'ask_size_max': ('ask_size', 'float'),
            'mid': ('mid', 'float'),
            'mid_min': ('mid', 'float'),
            'mid_max': ('mid', 'float'),
            'change_to_breakeven': ('change_to_breakeven', 'float'),
            'change_to_breakeven_min': ('change_to_breakeven', 'float'),
            'change_to_breakeven_max': ('change_to_breakeven', 'float'),
            'underlying_price': ('underlying_price', 'float'),
            'underlying_price_min': ('underlying_price', 'float'),
            'underlying_price_max': ('underlying_price', 'float'),
            'return_on_risk': ('return_on_risk', 'float'),
            'return_on_risk_min': ('return_on_risk', 'float'),
            'return_on_risk_max': ('return_on_risk', 'float'),
            'velocity': ('velocity', 'float'),
            'velocity_min': ('velocity', 'float'),
            'velocity_max': ('velocity', 'float'),
            'greeks_balance': ('greeks_balance', 'float'),
            'greeks_balance_min': ('greeks_balance', 'float'),
            'greeks_balance_max': ('greeks_balance', 'float'),
            'opp': ('opp', 'float'),
            'opp_min': ('opp', 'float'),
            'opp_max': ('opp', 'float'),
            'liquidity_score': ('liquidity_score', 'float'),
            'liquidity_score_min': ('liquidity_score', 'float'),
            'liquidity_score_max': ('liquidity_score', 'float')
            }

        
     
        # Dynamically build query based on kwargs
        query = f"SELECT * FROM poly_options WHERE oi > 0" if select_columns is None else f"SELECT ticker, strike, cp, expiry, {select_columns} FROM poly_options WHERE oi > 0"

        

        # Dynamically build query based on kwargs
        for key, value in kwargs.items():
            if key in column_types and value is not None:
                column, col_type = column_types[key]

                # Sanitize and format value for SQL query
                sanitized_value = self.sanitize_value(value, col_type)

                if 'min' in key:
                    query += f" AND {column} >= {sanitized_value}"
                elif 'max' in key:
                    query += f" AND {column} <= {sanitized_value}"
                else:
                    query += f" AND {column} = {sanitized_value}"
                print(query)
        

        try:
            async with self.pool.acquire() as conn:
                # Execute the query
                return await conn.fetch(query)
        except Exception as e:
            print(f"Error during query: {e}")
            return []
    async def fetch_filtered_records(self, **kwargs):
        # Start with the base SELECT statement
        select_sql = "SELECT ticker, strike, expiry, cp FROM poly_options"

        # Define which attributes can have min/max filters
        numeric_attrs = ['strike', 'dte', 'time_value', 'liquidity_score', 'theta', 
                         'theta_decay_rate', 'delta', 'delta_theta_ratio', 'gamma', 
                         'gamma_risk', 'vega', 'vega_impact', 'timestamp', 'oi', 'open', 
                         'high', 'low', 'close', 'intrinstic_value', 'extrinsic_value', 
                         'leverage_ratio', 'vwap', 'conditions', 'price', 'trade_size', 
                         'ask', 'bid', 'spread', 'spread_pct', 'iv', 'bid_size', 
                         'ask_size', 'vol', 'mid', 'change_to_breakeven', 
                         'underlying_price', 'return_on_risk', 'velocity', 'sensitivity', 
                         'greeks_balance', 'opp']

        # Add WHERE clauses based on kwargs
        conditions = []
        values = []
        for column, value in kwargs.items():
            print(kwargs)
            if column in numeric_attrs:
                if isinstance(value, dict):
                    # Handle min/max conditions for numeric attributes
                    for operation, op_value in value.items():
                        if operation == 'min':
                            conditions.append(f"{column} >= ${len(conditions) + 1}")
                            values.append(op_value)
                        elif operation == 'max':
                            conditions.append(f"{column} <= ${len(conditions) + 1}")
                            values.append(op_value)
                else:
                    # Handle exact match for numeric attributes
                    conditions.append(f"{column} = ${len(conditions) + 1}")
                    values.append(value)
            else:
                # Handle exact match for non-numeric attributes
                conditions.append(f"{column} = ${len(conditions) + 1}")
                values.append(value)

        if conditions:
            select_sql += " WHERE " + " AND ".join(conditions)

        try:
            async with self.pool.acquire() as conn:
                # Fetch records with the constructed query and values
                records = await conn.fetch(select_sql, *values)
                return records
        except asyncpg.PostgresError as e:
            print(f"Database error: {e}")
            return None
        except Exception as e:
            print(f"Unexpected error: {e}")
            return None

# Example usage

    async def fetch_records(self):
        select_sql = "SELECT * FROM poly_options"  # Modify this query based on your specific needs
        async with self.pool.acquire() as conn:
            # Fetch records
            records = await conn.fetch(select_sql)
            return records
    async def create_table(self, df, table_name, unique_column):
     
        print("Connected to the database.")
        dtype_mapping = {
            'int64': 'INTEGER',
            'float64': 'FLOAT',
            'object': 'TEXT',
            'bool': 'BOOLEAN',
            'datetime64': 'TIMESTAMP',
            'datetime64[ns]': 'timestamp',
            'datetime64[ms]': 'timestamp',
            'datetime64[ns, US/Eastern]': 'TIMESTAMP WITH TIME ZONE'
        }
    
        # Check for large integers and update dtype_mapping accordingly
        for col, dtype in zip(df.columns, df.dtypes):
            if dtype == 'int64':
                max_val = df[col].max()
                min_val = df[col].min()
                if max_val > 2**31 - 1 or min_val < -2**31:
                    dtype_mapping['int64'] = 'BIGINT'
        history_table_name = f"{table_name}_history"
        async with self.pool.acquire() as connection:

            table_exists = await connection.fetchval(f"SELECT to_regclass('{table_name}')")
            
            if table_exists is None:
                unique_constraint = f'UNIQUE ({unique_column})' if unique_column else ''
                create_query = f"""
                CREATE TABLE {table_name} (
                    {', '.join(f'"{col}" {dtype_mapping[str(dtype)]}' for col, dtype in zip(df.columns, df.dtypes))},
                    "insertion_timestamp" TIMESTAMP,
                    {unique_constraint}
                )
                """
                print(f"Creating table with query: {create_query}")

                # Create the history table
                history_create_query = f"""
                CREATE TABLE IF NOT EXISTS {history_table_name} (
                    operation CHAR(1) NOT NULL,
                    changed_at TIMESTAMP NOT NULL DEFAULT current_timestamp,
                    {', '.join(f'"{col}" {dtype_mapping[str(dtype)]}' for col, dtype in zip(df.columns, df.dtypes))}
                );
                """
                print(f"Creating history table with query: {history_create_query}")
                await connection.execute(history_create_query)
                try:
                    await connection.execute(create_query)
                    print(f"Table {table_name} created successfully.")
                except asyncpg.UniqueViolationError as e:
                    print(f"Unique violation error: {e}")
            else:
                print(f"Table {table_name} already exists.")
            
            # Create the trigger function
            trigger_function_query = f"""
            CREATE OR REPLACE FUNCTION save_to_{history_table_name}()
            RETURNS TRIGGER AS $$
            BEGIN
                INSERT INTO {history_table_name} (operation, changed_at, {', '.join(f'"{col}"' for col in df.columns)})
                VALUES (
                    CASE
                        WHEN (TG_OP = 'DELETE') THEN 'D'
                        WHEN (TG_OP = 'UPDATE') THEN 'U'
                        ELSE 'I'
                    END,
                    current_timestamp,
                    {', '.join('OLD.' + f'"{col}"' for col in df.columns)}
                );
                RETURN NEW;
            END;
            $$ LANGUAGE plpgsql;
            """
            await connection.execute(trigger_function_query)

            # Create the trigger
            trigger_query = f"""
            DROP TRIGGER IF EXISTS tr_{history_table_name} ON {table_name};
            CREATE TRIGGER tr_{history_table_name}
            AFTER UPDATE OR DELETE ON {table_name}
            FOR EACH ROW EXECUTE FUNCTION save_to_{history_table_name}();
            """
            await connection.execute(trigger_query)


            # Alter existing table to add any missing columns
            for col, dtype in zip(df.columns, df.dtypes):
                alter_query = f"""
                DO $$
                BEGIN
                    BEGIN
                        ALTER TABLE {table_name} ADD COLUMN "{col}" {dtype_mapping[str(dtype)]};
                    EXCEPTION
                        WHEN duplicate_column THEN
                        NULL;
                    END;
                END $$;
                """
                await connection.execute(alter_query)
    async def fetch(self, query):
        async with self.pool.acquire() as conn:
            records = await conn.fetch(query)
            return records

    async def option_trades(self, ticker:str,as_dataframe:bool=False):
        """
        Gets option trades for an option, or stock.

        >>> ticker: the ticker to query 


        >>> limit: the number of trades to return. (default 50000)
        

        >>> as_dataframe: whether to return as a dataframe or not (default false)
        """
        url = f"https://api.polygon.io/v3/trades/{ticker}?limit=50000&apiKey={self.api_key}"
       
        try:

            data = await self.paginate_concurrent(url, as_dataframe=True)
            if data is not None:
                return data
        except Exception as e:
            print(f'Error')

    async def rsi(self, ticker:str, timespan:str, limit:str='1000', window:int=14, date_from:str=None, date_to:str=None, session=None):
        """
        Arguments:

        >>> ticker

        >>> AVAILABLE TIMESPANS:

        minute
        hour
        day
        week
        month
        quarter
        year

        >>> date_from (optional) 
        >>> date_to (optional)
        >>> window: the RSI window (default 14)
        >>> limit: the number of N timespans to survey
        
        """

        if date_from is None:
            date_from = self.eight_days_ago

        if date_to is None:
            date_to = self.today


        endpoint = f"https://api.polygon.io/v1/indicators/rsi/{ticker}?timespan={timespan}&timestamp.gte={date_from}&timestamp.lte={date_to}&limit={limit}&window={window}&apiKey={self.api_key}"
 
        async with aiohttp.ClientSession() as session:
            try:
                async with session.get(endpoint) as resp:
                    datas = await resp.json()
                    if datas is not None:

                        
  
                
                        return RSI(datas, ticker)
            except (ClientConnectorError, ClientOSError, ContentTypeError):
                print(f"ERROR - {ticker}")
    async def rsi_snapshot(self, ticker:str):
        components = get_human_readable_string(ticker)
        symbol = components.get('underlying_symbol')
        strike = components.get('strike_price')
        expiry = components.get('expiry_date')
        call_put = components.get('call_put')
        call_put = str(call_put).lower()
        timespans = ['minute', 'hour', 'day', 'week', 'month']
        all_data_dicts=[]
        for timespan in timespans:
            rsi =await self.rsi(ticker, timespan, limit='1')
    

            # Check if rsi or rsi.rsi_value[0] is None
            # Check if rsi is not None and that rsi.rsi_value itself is not None and has at least one element
            if rsi is not None and rsi.rsi_value is not None and len(rsi.rsi_value) > 0:
                data_dict = { 
                    'timespan': timespan,
                    'option_symbol': ticker,
                    'ticker': symbol,
                    'strike': strike,
                    'expiry': expiry,
                    'call_put': call_put,
                    'rsi': rsi.rsi_value[0]
                }
            else:
                # Handle the None case appropriately
                # For example, you can set a default value or perform a different action
                data_dict = { 
                    'timespan': timespan,
                    'option_symbol': ticker,
                    'ticker': symbol,
                    'strike': strike,
                    'expiry': expiry,
                    'call_put': call_put,
                    'rsi': 0  # Replace 'default_value' with an appropriate fallback
                }
            all_data_dicts.append(data_dict)
        df = pd.DataFrame(all_data_dicts)
        
        return df

    async def option_aggregates(self, ticker:str, timespan:str='second', date_from:str='2019-09-17', date_to:str=None, limit=50000, 
    as_dataframe:bool=False):
        """Gets all aggregates for a ticker, or option symbol
        
        Arguments:

        >>> ticker: the ticker to query

        >>> timespan:

            -second
            -minute
            -hour
            -day
            -week
            -month
            -quarter
            -year

        >>> date_from: the date to start in YYYY-MM-DD format (default 2019-09-17)

        >>> date_to: the date to end on in YYYY-MM-DD format (default today)


        >>> limit: the number of results to return. (default 50000)


        >>> as_dataframe: whether to return as a dataframe or not (default false)

        """

        if date_to == None:
            date_to = self.today
        url = f"https://api.polygon.io/v2/aggs/ticker/{ticker}/range/1/{timespan}/{date_from}/{date_to}?adjusted=true&sort=desc&limit=50000&apiKey={self.api_key}"


       

        results = await self.paginate_concurrent(url, as_dataframe=False, concurrency=50)
        if as_dataframe == True:
            return pd.DataFrame(results)
        return results
            

    async def batch_insert_dataframe(self, df, table_name, unique_columns, batch_size=250):
        """
        WORKS - Creates table - inserts data based on DTYPES.
        
        """
     
        async with lock:
            if not await self.table_exists(table_name):
                await self.create_table(df, table_name, unique_columns)
            
            # Debug: Print DataFrame columns before modifications
            #print("Initial DataFrame columns:", df.columns.tolist())
            
            df = df.copy()
            df.dropna(inplace=True)
            df['insertion_timestamp'] = [datetime.now() for _ in range(len(df))]

            # Debug: Print DataFrame columns after modifications
            #print("Modified DataFrame columns:", df.columns.tolist())
            
            records = df.to_records(index=False)
            data = list(records)


            async with self.pool.acquire() as connection:
                column_types = await connection.fetch(
                    f"SELECT column_name, data_type FROM information_schema.columns WHERE table_name = '{table_name}'"
                )
                type_mapping = {col: next((item['data_type'] for item in column_types if item['column_name'] == col), None) for col in df.columns}

                async with connection.transaction():
                    insert_query = f"""
                    INSERT INTO {table_name} ({', '.join(f'"{col}"' for col in df.columns)}) 
                    VALUES ({', '.join('$' + str(i) for i in range(1, len(df.columns) + 1))})
                    ON CONFLICT ({unique_columns})
                    DO UPDATE SET {', '.join(f'"{col}" = excluded."{col}"' for col in df.columns)}
                    """
            
                    batch_data = []
                    for record in data:
                        new_record = []
                        for col, val in zip(df.columns, record):
                   
                            pg_type = type_mapping[col]

                            if val is None:
                                new_record.append(None)
                            elif pg_type == 'timestamp' and isinstance(val, np.datetime64):
                                new_record.append(pd.Timestamp(val).to_pydatetime().replace(tzinfo=None))

            
                            elif isinstance(val, datetime):
                                new_record.append(pd.Timestamp(val).to_pydatetime())
                            elif pg_type in ['timestamp', 'timestamp without time zone', 'timestamp with time zone'] and isinstance(val, np.datetime64):
                                new_record.append(pd.Timestamp(val).to_pydatetime().replace(tzinfo=None))  # Modified line
                            elif pg_type in ['double precision', 'real'] and not isinstance(val, str):
                                new_record.append(float(val))
                            elif isinstance(val, np.int64):  # Add this line to handle numpy.int64
                                new_record.append(int(val))
                            elif pg_type == 'integer' and not isinstance(val, int):
                                new_record.append(int(val))
                            else:
                                new_record.append(val)
                    
                        batch_data.append(new_record)

                        if len(batch_data) == batch_size:
                            try:
                                
                             
                                await connection.executemany(insert_query, batch_data)
                                batch_data.clear()
                            except Exception as e:
                                print(f"An error occurred while inserting the record: {e}")
                                await connection.execute('ROLLBACK')
                                raise

                if batch_data:  # Don't forget the last batch
       
                    try:

                        await connection.executemany(insert_query, batch_data)
                    except Exception as e:
                        print(f"An error occurred while inserting the record: {e}")
                        await connection.execute('ROLLBACK')
                        raise
    async def save_to_history(self, df, main_table_name, history_table_name):
        # Assume the DataFrame `df` contains the records to be archived
        if not await self.table_exists(history_table_name):
            await self.create_table(df, history_table_name, None)

        df['archived_at'] = datetime.now()  # Add an 'archived_at' timestamp
        await self.batch_insert_dataframe(df, history_table_name, None)
    async def table_exists(self, table_name):
        query = f"SELECT EXISTS (SELECT FROM information_schema.tables WHERE table_name = '{table_name}');"
  
        async with self.pool.acquire() as conn:
            async with conn.transaction():
                exists = await conn.fetchval(query)
        return exists
    async def save_structured_messages(self, data_list: list[dict], table_name: str):
        if not data_list:
            return  # No data to insert

        # Assuming all dicts in data_list have the same keys
        fields = ', '.join(data_list[0].keys())
        values_placeholder = ', '.join([f"${i+1}" for i in range(len(data_list[0]))])
        values = ', '.join([f"({values_placeholder})" for _ in data_list])
        
        query = f'INSERT INTO {table_name} ({fields}) VALUES {values}'
       
        async with self.pool.acquire() as conn:
            try:
                flattened_values = [value for item in data_list for value in item.values()]
                await conn.execute(query, *flattened_values)
            except UniqueViolationError:
                print('Duplicate - Skipping')


    async def paginate_concurrent(self, url, as_dataframe=False, concurrency=25):
        """
        Concurrently paginates through polygon.io endpoints that contain the "next_url"

        Arguments:

        >>> URL: the endpoint to query

        >>> as_dataframe: bool - return as a dataframe or not. (default false)

        >>> concurrency: int - the number of session to run (default 25)
        
        """

        all_results = []

        
        async with aiohttp.ClientSession() as session:
            pages_to_fetch = [url]
            
            while pages_to_fetch:
                tasks = []
                
                for _ in range(min(concurrency, len(pages_to_fetch))):
                    next_url = pages_to_fetch.pop(0)
                    tasks.append(self.fetch_page(next_url))
                    
                results = await asyncio.gather(*tasks)
                if results is not None:
                    for data in results:
                        if data is not None:
                            if "results" in data:
                                all_results.extend(data["results"])
                                
                            next_url = data.get("next_url")
                            if next_url:
                                next_url += f'&{urlencode({"apiKey": f"{self.api_key}"})}'
                                pages_to_fetch.append(next_url)
                        else:
                            break
        if as_dataframe:
            import pandas as pd
            return pd.DataFrame(all_results)
        else:
            return all_results
        



    #

    async def fetch_page(self, url):
        async with aiohttp.ClientSession() as session:
            async with session.get(url) as response:
                response.raise_for_status()
                return await response.json()
    

    async def fetch_endpoint(self, endpoint, params=None):
        """
        Uses endpoint / parameter combinations to fetch the polygon API

        dynamically.
        
        """
        # Build the query parameters
        filtered_params = {k: v for k, v in params.items() if v is not None} if params else {}
        
        if filtered_params:
            query_string = urlencode(filtered_params)
            if query_string:
                endpoint = f"{endpoint}?{query_string}&{self.api_key}"
                
        async with aiohttp.ClientSession() as session:
            async with session.get(url=endpoint, params=filtered_params) as response:
                response_data = await response.json()
                if 'next_url' in response_data:
                    # If "next_url" exists - the function auto paginates
                    return await self.paginate_concurrent(endpoint,as_dataframe=True,concurrency=40)
                return response_data  # or response.text() based on your needs
                
    async def get_price(self, ticker:str):
        """
        Fetches price from Webull API to use for option queries

        Arguments:

        >>> ticker: required - the ticker to survey
        """

        if ticker.startswith('I:'):
            ticker = ticker.replace('I:', '')
        datas = await trading.stock_quote(ticker)

        price = datas.web_stock_close

        return price
    async def get_option_strikes_by_moneyness(self, ticker, current_price, expiration_date):
        """
        Get option strikes categorized by moneyness and term.

        :param ticker: The ticker symbol of the underlying asset.
        :param current_price: Current price of the underlying asset.
        :return: Dictionary of options categorized by moneyness and term.
        """
        all_data = await self.get_option_chain_all(ticker, expiration_date=expiration_date)

        # Define your categorization thresholds for ITM, Deep ITM, ATM, OTM, Deep OTM
        deep_itm_threshold = 0.2  # e.g., 20% ITM is considered Deep ITM
        atm_threshold = 0.05      # e.g., within 5% of current price is considered ATM
        deep_otm_threshold = 0.2  # e.g., 20% OTM is considered Deep OTM

        # Define expiry terms (assuming all_data.df contains a 'expiry' column in YYYY-MM-DD format)
        today = datetime.today().date()
        one_month = today + timedelta(days=30)
        three_months = today + timedelta(days=90)

        near_term = lambda expiry_date: today <= expiry_date <= one_month
        mid_term = lambda expiry_date: one_month < expiry_date <= three_months
        far_term = lambda expiry_date: expiry_date > three_months

        categorized_strikes = {
            'Near Term': {'ITM': [], 'Deep ITM': [], 'ATM': [], 'OTM': [], 'Deep OTM': []},
            'Mid Term': {'ITM': [], 'Deep ITM': [], 'ATM': [], 'OTM': [], 'Deep OTM': []},
            'Far Term': {'ITM': [], 'Deep ITM': [], 'ATM': [], 'OTM': [], 'Deep OTM': []},
        }

        # Iterate through the options data and categorize based on moneyness and term
        for option in all_data.df.itertuples():
            # Calculate moneyness
            moneyness = (option.strike - current_price) / current_price
            expiry_date = datetime.strptime(option.expiry, '%Y-%m-%d').date()

            term_category = None
            if near_term(expiry_date):
                term_category = 'Near Term'
            elif mid_term(expiry_date):
                term_category = 'Mid Term'
            elif far_term(expiry_date):
                term_category = 'Far Term'

            moneyness_category = None
            if moneyness <= -deep_itm_threshold:
                moneyness_category = 'Deep ITM'
            elif -deep_itm_threshold < moneyness <= -atm_threshold:
                moneyness_category = 'ITM'
            elif -atm_threshold < moneyness < atm_threshold:
                moneyness_category = 'ATM'
            elif atm_threshold <= moneyness < deep_otm_threshold:
                moneyness_category = 'OTM'
            elif moneyness >= deep_otm_threshold:
                moneyness_category = 'Deep OTM'


            if term_category and moneyness_category:
                # Append a tuple of (option, moneyness_category) to the categorized strikes
                categorized_strikes[term_category][moneyness_category].append((option, moneyness_category))

        return categorized_strikes
    async def get_option_contracts(self, ticker:str, 
                                   strike_price_greater_than:str=None, 
                                   strike_price_less_than:str=None, 
                                   expiry_date_greater_than:str=None, 
                                   expiry_date_less_than:str=None, limit:str='250'):
        """
        Returns options for a specified ticker.

        Arguments:

        >>> ticker: the ticker to query (required)

        >>> strike_price_greater_than: the minimum strike to be returned (optional)      


        >>> strike_price_less_than: the maximum strike to be returned (optional) 


        >>> expiry_greater_than: the minimum expiry date to be returned (optional)


        >>> expiry_less_than: the maximum expiry date to be returned (optional)

        >>> limit: the amount of contracts to be returned (max 250 per request)

        
        """


        endpoint = f"https://api.polygon.io/v3/snapshot/options/{ticker}"

        
        params = {
            "limit": limit,
            "apiKey": self.api_key
        }

        df = await self.fetch_endpoint(endpoint, params)
      
        # Columns to normalize
        columns_to_normalize = ['day', 'underlying_asset', 'details', 'greeks', 'last_quote', 'last_trade']

        # Apply the helper function to each row and each specified column
        for column in columns_to_normalize:
            df = df.apply(lambda row: flatten_nested_dict(row, column), axis=1)

        return df

    async def get_strike_thresholds(self, ticker:str, price):
        indices_list = ["SPX", "SPXW", "NDX", "VIX", "VVIX"]
        if price is not None and ticker in indices_list:
            lower_strike = round(float(price) * 0.97)
            upper_strike = round(float(price) * 1.03)
            return lower_strike, upper_strike
        else:
            lower_strike = round(float(price) * 0.85)
            upper_strike = round(float(price) * 1.15)
            return lower_strike, upper_strike

    async def get_near_the_money_options(self, ticker: str, exp_greater_than:str=None, exp_less_than:str=None):
        """
        Gets options near the money for a ticker
        """

        if exp_greater_than is None:
            exp_greater_than = self.today

        if exp_less_than is None:
            exp_less_than = self.eight_days_from_now



        price = await self.get_price(ticker)
  
        if price is not None:
            upper_strike, lower_strike = await self.get_strike_thresholds(ticker, price)
      
            async with aiohttp.ClientSession() as session:
                url = f"https://api.polygon.io/v3/snapshot/options/{ticker}?strike_price.lte={lower_strike}&strike_price.gte={upper_strike}&expiration_date.gte={exp_greater_than}&expiration_date.lte={exp_less_than}&limit=250&apiKey={self.api_key}"
         
                async with session.get(url) as resp:
                    r = await resp.json()
                    results = r['results'] if 'results' in r else None
                    if results is None:
                        return
                    else:
                        results = UniversalOptionSnapshot(results)
                       
                        tickers = results.ticker
                        if ticker is not None:
                            atm_tickers = ','.join(tickers)
                            return atm_tickers
                        else:
                            return None
                        
    async def get_near_the_money_data(self, ticker: str, exp_greater_than:str='2023-11-17', exp_less_than:str='2023-12-15'):
        """
        Gets options near the money for a ticker
        """





        price = await self.get_price(ticker)
  
        if price is not None:
            upper_strike, lower_strike = await self.get_strike_thresholds(ticker, price)
      
            async with aiohttp.ClientSession() as session:
                url = f"https://api.polygon.io/v3/snapshot/options/{ticker}?strike_price.lte={lower_strike}&strike_price.gte={upper_strike}&expiration_date.gte={exp_greater_than}&expiration_date.lte={exp_less_than}&limit=250&apiKey={self.api_key}"
         
                data = await self.paginate_concurrent(url)

                return UniversalOptionSnapshot(data).df
            
    async def get_skew(self, ticker:str):
        """Gets the IV skew across all expiration dates along with the strikes above and below the lowest IV.
        
        Arguments:
        >>> ticker: required - the ticker to survey
        """

        options = await self.get_option_contracts(ticker, expiry_date_less_than='2023-11-03')

        # Sort by 'expiration_date' and 'implied_volatility'
        options_sorted = options.sort_values(by=['expiration_date', 'implied_volatility'])

        results = []

        # For each expiration date, get the lowest IV strike and the ones above and below it
        for expiration, group in options_sorted.groupby('expiration_date'):
            min_iv_idx = group['implied_volatility'].idxmin()
            
            # Get rows for lowest IV and the ones immediately above and below it
            rows = group.loc[[min_iv_idx - 1, min_iv_idx, min_iv_idx + 1], :]
            
            results.append(rows)

        # Concatenate all the results
        result_df = pd.concat(results)

     
        return result_df


    async def get_universal_snapshot(self, ticker): #✅
        """Fetches the Polygon.io universal snapshot API endpoint"""
        url=f"https://api.polygon.io/v3/snapshot?ticker.any_of={ticker}&limit=250&apiKey={os.environ.get('YOUR_POLYGON_KEY')}"
        print(url)
  
        async with aiohttp.ClientSession() as session:
            try:
                async with session.get(url) as resp:
                    data = await resp.json()
                    results = data['results'] if 'results' in data else None
                    if results is not None:
                        return UniversalOptionSnapshot(results)
                    else:
                        return None
            except ContentTypeError:
                pass


    async def get_option_chain_all(self, underlying_asset, strike_price=None, strike_price_lte=None, strike_price_gte=None, expiration_date=None, expiration_date_gte=None, expiration_date_lite=None, contract_type=None, order=None, limit=250, sort=None, pool=None):
        """
        Get all options contracts for an underlying ticker across all pages.

        :param underlying_asset: The underlying ticker symbol of the option contract.
        :param strike_price: Query by exact strike price of a contract.
        :param strike_price_lte: Query for strike price less than or equal to a value.
        :param strike_price_gte: Query for strike price greater than or equal to a value.
        :param expiration_date: Query by exact contract expiration with date format YYYY-MM-DD.
        :param expiration_date_gte: Query for expiration dates greater than or equal to a value.
        :param expiration_date_lite: Query for expiration dates less than or equal to a value.
        :param contract_type: Query by the type of contract (e.g., call or put).
        :param order: Order results based on the sort field (e.g., asc or desc).
        :param limit: Limit the number of results returned, default is 10 and max is 250.
        :param sort: Sort field used for ordering.
        :return: A list containing all option chain data across all pages.
        """
        params = {
            'strike_price': strike_price,
            'strike_price.lte': strike_price_lte,
            'strike_price.gte': strike_price_gte,
            'expiration_date': expiration_date,
            'expiration_date.gte': expiration_date_gte,
            'expiration_date.lte': expiration_date_lite,
            'contract_type': contract_type,
            'order': order,
            'limit': limit,
            'sort': sort
        }
        # Filter out None values
        params = {k: v for k, v in params.items() if v is not None}

        endpoint = f"https://api.polygon.io/v3/snapshot/options/{underlying_asset}"
        query_string = '&'.join([f"{k}={v}" for k, v in params.items()])
        if query_string:
            endpoint += '?' + query_string
        endpoint += f"&apiKey={self.api_key}"


        response_data = await self.paginate_concurrent(endpoint)
        option_data = UniversalOptionSnapshot(response_data)

        return option_data




    async def find_symbols(self, underlying_asset, strike_price=None, strike_price_lte=None, strike_price_gte=None, expiration_date=None, expiration_date_gte=None, expiration_date_lite=None, contract_type=None, order=None, limit=250, sort=None):
        """
        Get all options contracts for an underlying ticker across all pages.

        :param underlying_asset: The underlying ticker symbol of the option contract.
        :param strike_price: Query by exact strike price of a contract.
        :param strike_price_lte: Query for strike price less than or equal to a value.
        :param strike_price_gte: Query for strike price greater than or equal to a value.
        :param expiration_date: Query by exact contract expiration with date format YYYY-MM-DD.
        :param expiration_date_gte: Query for expiration dates greater than or equal to a value.
        :param expiration_date_lite: Query for expiration dates less than or equal to a value.
        :param contract_type: Query by the type of contract (e.g., call or put).
        :param order: Order results based on the sort field (e.g., asc or desc).
        :param limit: Limit the number of results returned, default is 10 and max is 250.
        :param sort: Sort field used for ordering.
        :return: A list containing all option chain data across all pages.
        """
        params = {
            'strike_price': strike_price,
            'strike_price.lte': strike_price_lte,
            'strike_price.gte': strike_price_gte,
            'expiration_date': expiration_date,
            'expiration_date.gte': expiration_date_gte,
            'expiration_date.lte': expiration_date_lite,
            'contract_type': contract_type,
            'order': order,
            'limit': limit,
            'sort': sort
        }
        # Filter out None values
        params = {k: v for k, v in params.items() if v is not None}

        endpoint = f"https://api.polygon.io/v3/snapshot/options/{underlying_asset}"
        query_string = '&'.join([f"{k}={v}" for k, v in params.items()])
        if query_string:
            endpoint += '?' + query_string
        endpoint += f"&apiKey={self.api_key}"


        response_data = await self.paginate_concurrent(endpoint)
        option_data = UniversalOptionSnapshot(response_data)

        return option_data
    

    async def insert_multiple_option_data(self, strike, expiry, dte, time_value, moneyness, liquidity_score, cp, 
                                          exercise_style, option_symbol, theta, theta_decay_rate, delta, 
                                          delta_theta_ratio, gamma, gamma_risk, vega, vega_impact, timestamp, 
                                          oi, open, high, low, close, intrinstic_value, extrinsic_value, 
                                          leverage_ratio, vwap, conditions, price, trade_size, exchange, ask, 
                                          bid, spread, spread_pct, iv, bid_size, ask_size, vol, mid, 
                                          change_to_breakeven, underlying_price, ticker, return_on_risk, 
                                          velocity):
        insert_sql = """
        INSERT INTO poly_opts (
            strike, expiry, dte, time_value, moneyness, liquidity_score, cp, 
            exercise_style, option_symbol, theta, theta_decay_rate, delta, 
            delta_theta_ratio, gamma, gamma_risk, vega, vega_impact, timestamp, 
            oi, open, high, low, close, intrinstic_value, extrinsic_value, 
            leverage_ratio, vwap, conditions, price, trade_size, exchange, ask, 
            bid, spread, spread_pct, iv, bid_size, ask_size, vol, mid, 
            change_to_breakeven, underlying_price, ticker, return_on_risk, 
            velocity
        ) VALUES (
            $1, $2, $3, $4, $5, $6, $7, $8, $9, $10, $11, $12, $13, $14, $15,
            $16, $17, $18, $19, $20, $21, $22, $23, $24, $25, $26, $27, $28,
            $29, $30, $31, $32, $33, $34, $35, $36, $37, $38, $39, $40, $41,
            $42, $43, $44
        )
        """
        data_tuples = list(zip(strike, expiry, dte, time_value, moneyness, liquidity_score, cp, 
                               exercise_style, option_symbol, theta, theta_decay_rate, delta, 
                               delta_theta_ratio, gamma, gamma_risk, vega, vega_impact, timestamp, 
                               oi, open, high, low, close, intrinstic_value, extrinsic_value, 
                               leverage_ratio, vwap, conditions, price, trade_size, exchange, ask, 
                               bid, spread, spread_pct, iv, bid_size, ask_size, vol, mid, 
                               change_to_breakeven, underlying_price, ticker, return_on_risk, 
                               velocity))

        async with self.pool.acquire() as conn:
            await conn.executemany(insert_sql, data_tuples)

    async def ensure_table_exists(self, conn, table_name, columns, column_types):
        # Combine column names and types into a list of strings
        column_definitions = ', '.join([f"{name} {dtype}" for name, dtype in zip(columns, column_types)])
        
        # Create a table if it does not exist (simple example, no primary key or other constraints)
        create_table_query = f"""
        CREATE TABLE IF NOT EXISTS {table_name} (
            {column_definitions}
        );
        """
        await conn.execute(create_table_query)



    async def yield_tickers(self):
        """
        An asynchronous generator that fetches and yields one ticker at a time from the database.

        Yields:
        str: A ticker from the database.
        """
        await self.connect()  # Ensure connection pool is ready
        async with self.pool.acquire() as conn:
            async with conn.transaction():
                # Define our SELECT query
                query = 'SELECT option_symbol FROM options_data'
                # Obtain a cursor
                cursor = await conn.cursor(query)
                # Manually iterate over the cursor
                while True:
                    record = await cursor.fetchrow()
                    if record is None:
                        break
                    ticker = record['option_symbol']
                    yield ticker

    async def get_tickers(self):
        tickers = []
        await self.connect()  # Ensure connection pool is ready
        async with self.pool.acquire() as conn:
            async with conn.transaction():
                query = 'SELECT option_symbol FROM options_data'
                cursor = await conn.cursor(query)
                # Fetch each record from the cursor
                record = await cursor.fetchrow()
                while record:
                    tickers.append(record['option_symbol'])
                    record = await cursor.fetchrow()
        return tickers

    async def fetch_iter(self, query, *params):
        async with self.pool.acquire() as conn:
            async with conn.transaction():
                async for record in conn.cursor(query):
                    yield record

    async def group_options_by_expiry(self, all_options: UniversalOptionSnapshot):
        """Takes all options and groups by expiration date."""
        


    async def chunk_250(self, all_options: UniversalOptionSnapshot):
        """Chunks all options into sizes of 250"""
        # Calculate the number of chunks
        total_items = len(all_options.ticker)  # Replace with the actual length of all_options
        chunk_size = 250
        num_chunks = (total_items + chunk_size - 1) // chunk_size  # Calculate the total number of chunks

        # Iterate over each chunk
        for i in range(num_chunks):
            start_index = i * chunk_size
            end_index = start_index + chunk_size
            chunk = all_options[start_index:end_index]  # Get the current chunk

            # Process or print the current chunk
            yield chunk

    async def full_skew(self, ticker:str):
        """
        Gets the lowest IV call/put for all expirations.

        Arguments:

        >>> ticker: the ticker to get the full skew for.

        """

        all_options = await self.get_option_chain_all(ticker)
        all_options_df = all_options.df

        # Empty list to store the lowest IV call and put options
        lowest_iv_options = []

        # Grouping by the 'expiry' column
        grouped = all_options_df.groupby('expiry')
        for expiry, group in grouped:
            # Filter call and put options
            calls = group[group['cp'] == 'call']
            puts = group[group['cp'] == 'put']

            # Combine calls and puts into one DataFrame
            combined_options = pd.concat([calls, puts])

            # Find the option (call or put) with the lowest IV
            lowest_iv_option = combined_options.loc[combined_options['iv'].idxmin()]

            # Add the lowest IV option to the list
            lowest_iv_options.append(lowest_iv_option)

        # Concatenate into a final DataFrame
        final_df = pd.DataFrame(lowest_iv_options)
        print(final_df.columns)
        columns_to_save = ['ticker', 'underlying_price', 'strike', 'expiry', 'dte', 'cp','gamma','vega','iv','vol','oi','ask', 'bid']
        # Define the list of columns you want to keep
        columns_to_keep = ['strike', 'expiry', 'cp', 'iv']  # Replace with your desired columns

        # Select only these columns from the DataFrame
        filtered_df = final_df[columns_to_keep]
        save_df = final_df[columns_to_save]

        save_df.to_csv('full_skew.csv', index=False)

        # Extract the constant values
        ticker = final_df['ticker'].iloc[0]  # Extracts the first value from the 'ticker' column
        underlying_price = final_df['underlying_price'].iloc[0]  # Extracts the first value from the 'underlying_price' column

        return filtered_df, underlying_price
    

    async def vol_oi_top_strike(self, ticker: str):
        """
        Gets the strikes with the highest volume and open interest across all expirations
        for both call and put options, returning specific columns.
        """

        all_options = await self.get_option_chain_all(ticker)
        all_options_df = all_options.df

        # Columns to return
        columns_to_return = ['ticker', 'strike', 'expiry', 'cp', 'oi', 'vol']

        # Create an empty DataFrame to store the final results
        final_df = pd.DataFrame()

        # Group by 'cp' (call/put) and 'expiry'
        grouped = all_options_df.groupby(['cp', 'expiry'])

        for (option_type, expiry), group in grouped:
            # Find the row with the max volume
            max_vol_row = group[group['vol'] == group['vol'].max()]

            # Find the row with the max open interest
            max_oi_row = group[group['oi'] == group['oi'].max()]

            # Append these rows to the final DataFrame
            # Selecting only the required columns
            final_df = pd.concat([final_df, max_vol_row[columns_to_return], max_oi_row[columns_to_return]])

        # Resetting the index of the final DataFrame
        final_df.reset_index(drop=True, inplace=True)

        return final_df

    async def lowest_theta(self, ticker: str):
        """
        Gets the option with the lowest theta for each expiry across all options,
        selecting the lowest out of the call or put for each expiry.
        """

        all_options = await self.get_option_chain_all(ticker)
        all_options_df = all_options.df

        # Ensure 'theta' is in numeric format
        all_options_df['theta'] = pd.to_numeric(all_options_df['theta'], errors='coerce')

        # Create an empty DataFrame to store the final results
        final_df = pd.DataFrame()

        # Group by 'expiry'
        grouped = all_options_df.groupby('expiry')

        for expiry, group in grouped:
            # Separate calls and puts
            calls = group[group['cp'] == 'call']
            puts = group[group['cp'] == 'put']

            # Find the row with the lowest theta for calls and puts
            min_theta_call = calls[calls['theta'] == calls['theta'].max()]
            min_theta_put = puts[puts['theta'] == puts['theta'].max()]

            # Select the one with the absolute lowest theta
            min_theta_option = pd.concat([min_theta_call, min_theta_put]).nsmallest(1, 'theta')

            # Append this row to the final DataFrame
            final_df = pd.concat([final_df, min_theta_option])

        # Resetting the index of the final DataFrame
        final_df.reset_index(drop=True, inplace=True)

        return final_df
    

    async def lowest_vega(self, ticker: str):
        """
        Gets the option with the lowest vega for each expiry across all options,
        selecting the lowest out of the call or put for each expiry.
        """

        all_options = await self.get_option_chain_all(ticker)
        all_options_df = all_options.df

        # Ensure 'theta' is in numeric format
        all_options_df['vega'] = pd.to_numeric(all_options_df['vega'], errors='coerce')

        # Create an empty DataFrame to store the final results
        final_df = pd.DataFrame()

        # Group by 'expiry'
        grouped = all_options_df.groupby('expiry')

        for expiry, group in grouped:
            # Separate calls and puts
            calls = group[group['cp'] == 'call']
            puts = group[group['cp'] == 'put']

            # Find the row with the lowest theta for calls and puts
            min_theta_call = calls[calls['vega'] == calls['vega'].max()]
            min_theta_put = puts[puts['vega'] == puts['vega'].max()]

            # Select the one with the absolute lowest theta
            min_theta_option = pd.concat([min_theta_call, min_theta_put]).nsmallest(1, 'vega')

            # Append this row to the final DataFrame
            final_df = pd.concat([final_df, min_theta_option])

        # Resetting the index of the final DataFrame
        final_df.reset_index(drop=True, inplace=True)

        return final_df

    async def highest_velocity(self, ticker: str):
        """
        Gets the option with the highest velocity for each expiry across all options,
        selecting the highest out of the call or put for each expiry.
        """

        all_options = await self.get_option_chain_all(ticker)
        all_options_df = all_options.df

        # Ensure 'velocity' is in numeric format
        all_options_df['velocity'] = pd.to_numeric(all_options_df['velocity'], errors='coerce')

        # Create an empty DataFrame to store the final results
        final_df = pd.DataFrame()

        # Group by 'expiry'
        grouped = all_options_df.groupby('expiry')

        for expiry, group in grouped:
            # Separate calls and puts
            calls = group[group['cp'] == 'call']
            puts = group[group['cp'] == 'put']

            # Find the row with the highest velocity for calls and puts
            max_velocity_call = calls[calls['velocity'] == calls['velocity'].max()]
            max_velocity_put = puts[puts['velocity'] == puts['velocity'].max()]

            # Select the one with the absolute highest velocity
            max_velocity_option = pd.concat([max_velocity_call, max_velocity_put]).nlargest(1, 'velocity')

            # Append this row to the final DataFrame
            final_df = pd.concat([final_df, max_velocity_option])

        # Resetting the index of the final DataFrame
        final_df.reset_index(drop=True, inplace=True)

        return final_df


    def binomial_american_option(self, S, K, T, r, sigma, N, option_type='put'):
        """
        Binomial model for American option pricing.

        :param S: Current stock price.
        :param K: Strike price.
        :param T: Time to expiry in years.
        :param r: Risk-free interest rate.
        :param sigma: Volatility of the stock.
        :param N: Number of steps in the binomial tree.
        :param option_type: Type of the option ('call' or 'put').
        :return: Estimated option price.
        """
        dt = T / N  # Time step
        u = np.exp(sigma * np.sqrt(dt))  # Upward movement factor
        d = 1 / u  # Downward movement factor
        p = (np.exp(r * dt) - d) / (u - d)  # Risk-neutral probability

        # Initialize asset price at maturity
        ST = np.zeros((N+1, N+1))
        ST[0, 0] = S

        # Create the binomial tree
        for i in range(1, N + 1):
            ST[i, 0] = ST[i - 1, 0] * u
            for j in range(1, i + 1):
                ST[i, j] = ST[i - 1, j - 1] * d

        # Initialize option values at maturity
        option = np.zeros((N+1, N+1))
        for j in range(N+1):
            if option_type == 'call':
                option[N, j] = max(0, ST[N, j] - K)
            else:
                option[N, j] = max(0, K - ST[N, j])

        # Step backwards through the tree
        for i in range(N-1, -1, -1):
            for j in range(i + 1):
                option_early_exercise = max(ST[i, j] - K, 0) if option_type == 'call' else max(K - ST[i, j], 0)
                option[i, j] = max(option_early_exercise, np.exp(-r * dt) * (p * option[i + 1, j] + (1 - p) * option[i + 1, j + 1]))

        return option[0, 0]

    async def get_theoretical_price(self, ticker):
        # Fetching necessary data
        prices = self.yf.fast_info(ticker)
        current_price = prices[prices[0] == 'lastPrice'][1].values[0]
    
        all_options_data = await self.get_option_chain_all(underlying_asset=ticker)

        risk_free_rate = 0.0565  # Current risk-free rate (5.65%)
        N = 100  # Number of steps in the binomial tree

        theoretical_values = []

    

        for bid, ask, iv, strike, expiry, option_type, dtes, volume, oi, velocity, delta, vega, gamma, theta, close in zip(
                all_options_data.bid, all_options_data.ask, all_options_data.implied_volatility, 
                all_options_data.strike, all_options_data.expiry, all_options_data.contract_type, all_options_data.days_to_expiry, all_options_data.open_interest, all_options_data.volume, all_options_data.option_velocity, all_options_data.delta, all_options_data.vega, all_options_data.gamma, all_options_data.theta, all_options_data.underlying_price):
            def convert_days_to_years(days):
                # Assuming 365 days in a year
                return float(days) / 365.0

            T = convert_days_to_years(dtes)
            
            # Check if implied volatility is None
            if iv is None:
                # Handle the case where implied volatility is not available.
                # You might want to skip this option, use an average IV, or another method.
                continue
            if delta is None:
                continue
            if gamma is None:
                continue
            if theta is None:
                continue
            if vega is None:
                continue

            sigma = iv  # Implied volatility

            # Calculate theoretical price using the binomial model
            theoretical_price = self.binomial_american_option(current_price, strike, T, risk_free_rate, sigma, N, option_type)
            
            theoretical_values.append({
                'ticker': ticker,
                'current_price': close,
                'iv': iv,
                'velocity': velocity,
                'strike': strike,
                'expiry': expiry,
                'bid': bid,
                'theoretical_price': theoretical_price,
                'ask': ask,
                'volume': volume,
                'oi': oi,
                'delta': delta,
                'gamma': gamma,
                'vega': vega,
                'theta': theta,
                'type': option_type
            })

        return theoretical_values
    async def gpt_filter(self, ticker:str=None, strike:float=None, strike_min:float=None, strike_max:float=None, cp:str=None, expiry:str=None, expiry_min:str=None, expiry_max:str=None, dte_max:int=None, dte_min:int=None, 
                liquidity_score_min:float=None, liquidity_score_max:float=None, theta_min:float=None, theta_max:float=None, 
                delta_min:float=None, delta_max:float=None, gamma_min:float=None, gamma_max:float=None, vega_min:float=None, 
                vega_max:float=None, timestamp_min:float=None, timestamp_max:float=None, oi_min:float=None, oi_max:float=None, 
                open_min:float=None, open_max:float=None, high_min:float=None, high_max:float=None, low_min:float=None, 
                low_max:float=None, close_min:float=None, close_max:float=None, intrinsic_value_min:float=None, 
                intrinsic_value_max:float=None, extrinsic_value_min:float=None, extrinsic_value_max:float=None, 
                leverage_ratio_min:float=None, leverage_ratio_max:float=None, vwap_min:float=None, vwap_max:float=None, 
                price_min:float=None, price_max:float=None, trade_size_min:float=None, trade_size_max:float=None, 
                ask_min:float=None, ask_max:float=None, bid_min:float=None, bid_max:float=None, spread_min:float=None, 
                spread_max:float=None, iv_min:float=None, iv_max:float=None, mid_min:float=None, mid_max:float=None, 
                underlying_price_min:float=None, underlying_price_max:float=None, return_on_risk_min:float=None, 
                return_on_risk_max:float=None, velocity_min:float=None, velocity_max:float=None, sensitivity_min:float=None, 
                sensitivity_max:float=None, greeks_balance_min:float=None, greeks_balance_max:float=None, 
                opp_min:float=None, opp_max:float=None, exchange:str=None, bid_size_min:float=None, bid_size_max:float=None, ask_size_min:float=None, ask_size_max:float=None, vol_min:float=None, vol_max:float=None):
        """Returns table chunks of filtered options!"""
        await self.connect()

        # Manually construct the params dictionary
        params = locals()
        params = {k: v for k, v in params.items() if v is not None and k != 'opts'}

        records = await self.filter_options(**params)

        # Process and print the fetched records
        df = pd.DataFrame(records)
        table = tabulate(df, headers=['sym', 'strike', 'cp', 'expiry', 'theta'], tablefmt='fancy', showindex=False)
        # Break apart data into chunks of 4000 characters
        chunks = chunk_string(table, 4000)
        # Fetch records

        return chunks
    async def close(self):
        if self.pool:
            await self.pool.close()
            self.pool = None



import os


from dotenv import load_dotenv
load_dotenv()
import psycopg2



class DiscordDBManager:
    def __init__(self):
            


        self.channel_types = { 

            'category': '4',
            
            
        }

        self.member_permissions = '71453223935041'

        self.role_dict = {
            'LIFETIME MEMBER': 1002249878283493456,
            'youtube support': 1086118401660964914,
            'youtube level1': 1086118401660964916,
            'youtube level2': 1086118401660964917,
            'patreon level1': 896207245853999145,
            'patreon level2': 941029523699400705,
            'patreon level3': 938824920589283348,
            'patreon last': 1145204112481321040

        }
       
        self.fudstop_id = 888488311927242753 # replace with your guild ID
        self.test_id=1193717419328409600
        self.token = os.environ.get('DISCORD_AUTHORIZATION')
        print(self.token)
        self.headers = {
                "authority": "discord.com",
                "method": "POST",
                "path": "/api/v9/science",
                "scheme": "https",
                "Accept": "*/*",
                'Content-Type': 'application/json',
                "Accept-Encoding": "gzip, deflate, br",
                "Accept-Language": "en-US,en;q=0.9",
                "Authorization": self.token,
                "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/120.0.0.0 Safari/537.36",

        }

    def create_thread_table(self, connection):
        """
        Creates a PostgreSQL database table to store thread-related information.

        This function defines the structure of the "thread" table with columns
        for thread attributes such as thread_id, parent_channel, name, total_messages,
        message_count, last_message_id, and guild.

        The table is created if it doesn't already exist, and a unique constraint
        is added on the "thread_id" column. This ensures that if the same thread_id
        is inserted again, it will update the existing row with the new values.

        :return: None
        """
        try:

            if connection:
                cursor = connection.cursor()

                # Define the SQL CREATE TABLE statement for the thread table
                create_table_query = """
                    CREATE TABLE IF NOT EXISTS threads (
                        id SERIAL PRIMARY KEY,
                        thread_id BIGINT UNIQUE,
                        parent_channel BIGINT,
                        name TEXT,
                        total_messages INT,
                        message_count INT,
                        last_message_id BIGINT,
                        guild BIGINT
                    );
                """

                # Execute the CREATE TABLE statement
                cursor.execute(create_table_query)
                connection.commit()
                cursor.close()
                print("Thread table created successfully")
        except (Exception, psycopg2.Error) as error:
            print("Error creating the thread table:", error)
        finally:
            if connection:
                connection.close()

    def insert_thread(self, thread_id, parent_channel, name, total_messages, message_count, last_message_id, guild):
        try:
            connection = self.connect_to_database()
            if connection:
                cursor = connection.cursor()
                # Define the SQL INSERT statement
                insert_query = """
                    INSERT INTO threads (thread_id, parent_channel, name, total_messages, message_count, last_message_id, guild)
                    VALUES (%s, %s, %s, %s, %s, %s, %s);
                """
                # Execute the INSERT statement with the retrieved attributes
                cursor.execute(insert_query, (thread_id, parent_channel, name, total_messages, message_count, last_message_id, guild))
                connection.commit()
                cursor.close()
                print("Insertion successful")
        except (Exception, psycopg2.Error) as error:
            print("Error inserting data into the database:", error)
        finally:
            if connection:
                connection.close()
    


import disnake

from disnake.ext import commands
from apis.gexbot.gexbot import GexMajorLevels,GEXBot,Gex,MaxGex
from _markets.list_sets.ticker_lists import gex_tickers
import asyncio



# At the top of the file.
import disnake
from disnake.ext import commands





class GEXCog(commands.Cog):
    def __init__(self, bot):
        self.bot=bot
        self.tickers = gex_tickers
        self.gexbot = GEXBot()




    @commands.slash_command()
    async def gex(self, inter):
        pass




    @gex.sub_command()
    async def get_gex(self, inter:disnake.AppCmdInter, ticker:str=commands.Param(choices=gex_tickers)):
        """Gets GEX data for a ticker - Work in progress"""
        ticker = ticker.upper()
        await inter.response.defer()
        while True:

            gex = await self.gexbot.get_gex(ticker)


            await inter.edit_original_message(f"> **{gex}**")

            await asyncio.sleep(5)



    @gex.sub_command()
    async def gex_major_levels(self, inter:disnake.AppCmdInter, ticker:str=commands.Param(choices=gex_tickers)):
        """Returns major GEX Levels"""
        ticker = ticker.upper()
        await inter.response.defer()
        counter = 0
        while True:
            counter = counter + 1
            data = await self.gexbot.major_levels(ticker)
            embed = disnake.Embed(title=f"Gex Major Levels - {ticker}", description=f"```py\n{data.as_dataframe}```", color=disnake.Colour.dark_gold())
        
            embed.set_footer(text='Provided by KRAKENSLAYER')


            await inter.edit_original_message(embed=embed)

            if counter == 250:
                await inter.edit_original_message(f"> Stream ended. Use /gex_major_levels to run again.")

    @gex.sub_command()
    async def gex_spy_spx(self, inter:disnake.AppCmdInter):

        """Comparitive GEX between SPY/SPX"""
    
        await inter.response.defer()
        counter = 0
        while True:
            counter = counter + 1
            data = await self.gexbot.major_levels('SPX')
            data2 = await self.gexbot.major_levels('SPY')
            embed = disnake.Embed(title=f"Gex Major Levels - SPY/SPX", color=disnake.Colour.dark_gold())
            
            embed.add_field(name=f"Zero Gamma:", value=f"> SPY: **{round(float(data2.zero_gamma),4)}**\n> SPX: **{round(float(data.zero_gamma),2)}**")
            embed.add_field(name=f"Major OI:", value=f"> SPY POS: **{data2.mpos_oi}**\n> SPY NEG: **{data2.mneg_oi}**\n\n> SPX POS: **{data.mpos_oi}**\n> SPX NEG: **{data.mneg_oi}**")
            embed.add_field(name=f"Major Vol:", value=f"> SPY POS: **{data2.mpos_vol}**\n> SPY NEG: **{data2.mneg_vol}**\n\n> SPX POS: **{data.mpos_vol}**\n> SPX NEG: **{data.mneg_vol}**")
            embed.add_field(name=f"Prices:", value=f"> SPY: **${data2.spot}**\n> SPX: **${data.spot}**")
            embed.set_footer(text='Provided by KRAKENSLAYER')


            await inter.edit_original_message(embed=embed)

            if counter == 250:
                await inter.edit_original_message(f"> Stream ended. Use /gex_spy_spx to run again.")
                break

    async def run_gex(self):
    
        tasks = [self.gex(i) for i in gex_tickers]
        await asyncio.gather(*tasks)


gex = GEXCog(commands.Bot)




class GexSelect(disnake.ui.Select):
    def __init__(self, ticker=None):
        self.ticker=ticker
        super().__init__( 
            placeholder='Select A Function -->',
            min_values=1,
            max_values=1,
            custom_id='gexSelect',
            options= [ 
                disnake.SelectOption(label='Get Gex', value='0', description=f'Get GEX data for a ticker.'),
                disnake.SelectOption(label='GEX Major Levels', value='1', description=f'Get major GEX levels for a ticker.'),
                disnake.SelectOption(label='SPY // SPX', value='2', description='Get comparitive GEX between SPY & SPX')
            ]
        )


    async def callback(self, inter:disnake.AppCmdInter):
        if self.values[0] == '0':
            await gex.get_gex(inter, self.ticker)

        elif self.values[0] == '1':
            await gex.gex_major_levels(inter, self.ticker)


        elif self.values[0] == '2':
            await gex.gex_spy_spx(inter)


def setup(bot: commands.Bot):
    bot.add_cog(GEXCog(bot))
    print(f"Gex commands - READY!")


import os
from dotenv import load_dotenv
from datetime import datetime, timedelta
import pandas as pd
import requests
from .webull_options import WebullOptionsData
from typing import List
from .screener_models import ScreenerRule, TickerInfo
load_dotenv()




class WebulScreener:
    def __init__(self):
        self.pool = None
        self.as_dataframe = None
        self.session = requests.session()
        self.account_id=os.environ.get('WEBULL_ACCOUNT_ID')
        self.api_key = os.environ.get('YOUR_POLYGON_KEY')
        self.today = datetime.now().strftime('%Y-%m-%d')
        self.yesterday = (datetime.now() - timedelta(days=1)).strftime('%Y-%m-%d')
        self.tomorrow = (datetime.now() + timedelta(days=1)).strftime('%Y-%m-%d')
        self.thirty_days_ago = (datetime.now() - timedelta(days=30)).strftime('%Y-%m-%d')
        self.thirty_days_from_now = (datetime.now() + timedelta(days=30)).strftime('%Y-%m-%d')
        self.fifteen_days_ago = (datetime.now() - timedelta(days=15)).strftime('%Y-%m-%d')
        self.fifteen_days_from_now = (datetime.now() + timedelta(days=15)).strftime('%Y-%m-%d')
        self.eight_days_from_now = (datetime.now() + timedelta(days=8)).strftime('%Y-%m-%d')
        self.eight_days_ago = (datetime.now() - timedelta(days=8)).strftime('%Y-%m-%d')
        self.headers = {
        "Access_token": os.environ.get('webull_access_token'),
        "Accept": "*/*",
        "App": "global",
        "App-Group": "broker",
        "Appid": "wb_web_app",
        "Content-Type": "application/json",
        "Device-Type": "Web",
        "Did": os.environ.get('DID'),
        "Hl": "en",
        "Locale": "eng",
        "Os": "web",
        "Osv": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/119.0.0.0 Safari/537.36",
        "Ph": "Windows Chrome",
        "Platform": "web",
        "Referer": "https://app.webull.com/",
        "Reqid": os.environ.get('REQ_ID'),
        "Sec-Ch-Ua": "\"Chromium\";v=\"118\", \"Google Chrome\";v=\"118\", \"Not=A?Brand\";v=\"99\"",
        "Sec-Ch-Ua-Mobile": "?0",
        "Sec-Ch-Ua-Platform": "\"Windows\"",
        "T_time": "1698276695206",
        "Tz": "America/Los_Angeles",
        "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/118.0.0.0 Safari/537.36",
        "Ver": "4.0.6",
        "X-S": os.environ.get('YOUR_X-S'),
        "X-Sv": "xodp2vg9"
    }

    def get_option_data_for_id(self, id, symbol):
        print(f"Starting processing for ticker: {id}")
        dataframes = []  # Initialize a list to collect DataFrames
  

        print(f"Processing batch ID: {id} for ticker: {id}")
        url = f"https://quotes-gw.webullfintech.com/api/quote/option/quotes/queryBatch?derivativeIds={id}"

        data = self.session.get(url, headers=self.headers).json()
        if not data:  # If data is empty or None, break the loop
            print(f"No more data for ticker: {id}. Moving to next.")
          
        wb_data = WebullOptionsData(data)
        if self.as_dataframe is not None:
            df = wb_data.as_dataframe
            df['ticker'] = id
            df['symbol'] = symbol
            df = df.rename(columns={'open_interest_change': 'oi_change'})
            return df
    def query(self, 
            ask_gte=None, ask_lte=None,
            bid_gte=None, bid_lte=None,
            changeRatio_gte=None, changeRatio_lte=None,
            close_gte=None, close_lte=None,
            delta_gte=None, delta_lte=None,
            direction=None,
            expireDate_gte=None, expireDate_lte=None,
            gamma_gte=None, gamma_lte=None,
            implVol_gte=None, implVol_lte=None,
            openInterest_gte=None, openInterest_lte=None,
            rho_gte=None, rho_lte=None,
            theta_gte=None, theta_lte=None,
            vega_gte=None, vega_lte=None,
            volume_gte=None, volume_lte=None):

        url = "https://quotes-gw.webullfintech.com/api/wlas/option/screener/query"
        # Initialize an empty dictionary for the filter
        filter_dict = {}

        # Helper function to add filter criteria
        def add_filter_criteria(name, gte, lte):
            if gte is not None and lte is not None:
                filter_dict[name] = f"gte={gte}&lte={lte}"

        # Add criteria to the filter dictionary if they are provided
        add_filter_criteria("options.screener.rule.ask", ask_gte, ask_lte)
        add_filter_criteria("options.screener.rule.bid", bid_gte, bid_lte)
        add_filter_criteria("options.screener.rule.changeRatio", changeRatio_gte, changeRatio_lte)
        add_filter_criteria("options.screener.rule.close", close_gte, close_lte)
        add_filter_criteria("options.screener.rule.delta", delta_gte, delta_lte)
        if direction:
            filter_dict["options.screener.rule.direction"] = direction
        add_filter_criteria("options.screener.rule.expireDate", expireDate_gte, expireDate_lte)
        add_filter_criteria("options.screener.rule.gamma", gamma_gte, gamma_lte)
        add_filter_criteria("options.screener.rule.implVol", implVol_gte, implVol_lte)
        add_filter_criteria("options.screener.rule.openInterest", openInterest_gte, openInterest_lte)
        add_filter_criteria("options.screener.rule.rho", rho_gte, rho_lte)
        add_filter_criteria("options.screener.rule.theta", theta_gte, theta_lte)
        add_filter_criteria("options.screener.rule.vega", vega_gte, vega_lte)
        add_filter_criteria("options.screener.rule.volume", volume_gte, volume_lte)

        # Build the payload
        payload = {
            "filter": filter_dict,
            "page": {'fetchSize': 25}
        }

        
        data= self.session.post(url, headers=self.headers, json=payload).json()
        datas = data.get('datas')

        belongTickerId = [i.get('belongTickerId') for i in datas]
        derivativeId = [i.get('derivativeId') for i in datas]
        derivative = [i.get('derivative') for i in datas]
        values = [i.get('values') for i in datas]

            


        derivative = TickerInfo(derivative)
    
        data_dict = { 

            'symbol': derivative.unSymbol,
            'strike': derivative.strikePrice,
            'call_put': derivative.direction,
            'expiry': derivative.expireDate,
            'id': derivativeId
       
        }

    

        return data_dict


import disnake
class ScreenerSelect(disnake.ui.Select):
    def __init__(self, data_dict):
        self.data_dict = data_dict

        options = []
        # Ensure all lists are of the same length
        num_items = len(data_dict['id'])
        if all(len(data_dict[key]) == num_items for key in ['symbol', 'strike', 'expiry', 'call_put']):
            for i in range(num_items):
                # Construct the label and value for each option
                label = f"{data_dict['symbol'][i]} | Strike: {data_dict['strike'][i]} | Expiry: {data_dict['expiry'][i]} | {data_dict['call_put'][i]}"
                value = str(data_dict['id'][i])

                options.append(disnake.SelectOption(label=label, value=value))
        else:
            print("Data lists are not of equal length")

        super().__init__(
            placeholder="> Results >",
            min_values=1,
            max_values=len(options),
            custom_id='optionscreener',
            options=options
        )
        super().__init__(
            placeholder="> Results >",
            min_values=1,
            max_values=len(data_dict),
            custom_id='optionscreener',
            options=options
        )




    async def callback(self, inter:disnake.AppCmdInter):
        await inter.response.defer()
        if self.values[0]:

            url = requests.get(f"https://quotes-gw.webullfintech.com/api/quote/option/quotes/queryBatch?derivativeIds={','.join(self._selected_values)}").json()
            print(url)

            data = WebullOptionsData(url).data_dict

            description = self.format_data_for_embed(data)
            embed = disnake.Embed(title=f"Results for Option:", description=f"> ***{description}***")
            view = disnake.ui.View()
            await inter.edit_original_message(embed=embed)

    def format_data_for_embed(self, data):
        # Format each key-value pair in data
        formatted_data = []
        for key, values in data.items():
            formatted_values = ', '.join(str(value) for value in values)
            formatted_data.append(f"**{key}:** {formatted_values}")

        # Join all formatted data into a single string
        return '\n'.join(formatted_data)